__all__ = ('Preinstance',)

from scarletio import RichAttributeErrorBaseType


class Preinstance(RichAttributeErrorBaseType):
    """
    An object that is picked up and created an instance with after the type is initialised.
    
    Attributes
    ----------
    keyword_parameters : `dict<str, object>`
        Additional keyword parameters to preinstance with.
    
    name : `str`
        The instance's name.
    
    positional_parameters : `tuple<object>`
        Additional positional parameters to preinstance with.
    
    value : `int | str`
        The instance's value.
    """
    __slots__ = ('keyword_parameters', 'name', 'positional_parameters', 'value')
    
    def __new__(cls, value, name, *positional_parameters, **keyword_parameters):
        """
        Creates a new value to preinstance..
        
        Parameters
        ----------
        value : `int | str`
            The instance's value.
        
        name : `str`
            The instance's name.
        
        *positional_parameters : Positional parameters
            Additional positional parameters.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        """
        self = object.__new__(cls)
        self.keyword_parameters = keyword_parameters
        self.name = name
        self.positional_parameters = positional_parameters
        self.value = value
        return self
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = [type(self).__name__, '(']
        
        # value
        repr_parts.append(repr(self.value))
        
        # name
        repr_parts.append(', ')
        repr_parts.append(repr(self.name))
        
        # positional_parameters
        for positional_parameter in self.positional_parameters:
            repr_parts.append(', ')
            repr_parts.append(repr(positional_parameter))
        
        # keyword_parameters
        keyword_parameters = self.keyword_parameters
        if keyword_parameters:
            for key, value in sorted(keyword_parameters.items()):
                repr_parts.append(', ')
                repr_parts.append(key)
                repr_parts.append(' = ')
                repr_parts.append(repr(value))
        
        repr_parts.append(')')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = 0
        
        # value
        hash_value ^= hash(self.value)
        
        # name
        hash_value ^= hash(self.name)
        
        # positional_parameters
        hash_value ^= hash(self.positional_parameters)
        
        # keyword_parameters
        keyword_parameters = self.keyword_parameters
        if keyword_parameters:
            hash_value ^= hash(tuple(sorted(keyword_parameters)))
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # value
        if self.value != other.value:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # positional_parameters
        if self.positional_parameters != other.positional_parameters:
            return False
        
        # keyword_parameters
        if self.keyword_parameters != other.keyword_parameters:
            return False
        
        return True



__all__ = ('create_partial_role_from_id', 'parse_role', 'parse_role_mention')

from functools import partial as partial_func

from scarletio import export

from ...core import ROLES
from ...utils import ID_RP, ROLE_MENTION_RP

from .fields import (
    put_color, put_flags, put_mentionable, put_name, put_permissions, put_position,
    put_separated, put_unicode_emoji, validate_color, validate_flags, validate_mentionable, validate_name,
    validate_permissions, validate_position, validate_separated, validate_unicode_emoji
)
from .role import ROLE_ICON, Role


ROLE_FIELD_CONVERTERS = {
    'color': (validate_color, put_color),
    'flags': (validate_flags, put_flags),
    'icon': (partial_func(ROLE_ICON.validate_icon, allow_data = True), partial_func(ROLE_ICON.put_into, as_data = True)),
    'mentionable': (validate_mentionable, put_mentionable),
    'name': (validate_name, put_name),
    'permissions': (validate_permissions, put_permissions),
    'position': (validate_position, put_position),
    'separated': (validate_separated, put_separated),
    'unicode_emoji': (validate_unicode_emoji, put_unicode_emoji),
}


@export
def create_partial_role_from_id(role_id, guild_id = 0):
    """
    Creates a partial role from the given `role_id`. If the role already exists returns that instead.
    
    Parameters
    ----------
    role_id : `int`
        The unique identifier number of the role.
    guild_id : `int` = `0`, Optional
        The role's guild's identifier.
    
    Returns
    -------
    role : ``Role``
    """
    try:
        return ROLES[role_id]
    except KeyError:
        pass
    
    role = Role._create_empty(role_id, guild_id)
    ROLES[role_id] = role
    
    return role


def parse_role_mention(text):
    """
    If the text is a role mention, returns the respective role if found.
    
    Parameters
    ----------
    text : `str`
       The text to parse the role out from.
    
    Returns
    -------
    role : `None`, ``Role``
        The found role if any.
    """
    parsed = ROLE_MENTION_RP.fullmatch(text)
    if parsed is None:
        return
    
    role_id = int(parsed.group(1))
    return ROLES.get(role_id, None)


def parse_role(text, guild = None):
    """
    Tries to parse a role out from the given text.
    
    Parameters
    ----------
    text : `str`
        The text to parse the role out.
    
    guild : `None`, ``Guild`` = `None`, Optional
        Context for name based parsing.
    
    Returns
    -------
    role : `None`, ``Role``
        The found role if any.
    """
    parsed = ID_RP.fullmatch(text)
    if (parsed is not None):
        role_id = int(parsed.group(1))
        try:
            role = ROLES[role_id]
        except KeyError:
            pass
        else:
            return role
    
    role = parse_role_mention(text)
    if (role is not None):
        return role
    
    if (guild is not None):
        if (guild is not None):
            role = guild.get_role_like(text)
            if (role is not None):
                return role
    
    return None



__all__ = ()

from ...field_validators import nullable_date_time_validator_factory
from ...utils import datetime_to_millisecond_unix_time, millisecond_unix_time_to_datetime

# end

def parse_end(data):
    """
    Parses activity timestamp end time from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Activity timestamp data.
    
    Returns
    -------
    end : `None`, `datetime`
    """
    end = data.get('end', None)
    if (end is not None):
        return millisecond_unix_time_to_datetime(end)


def put_end(end, data, defaults):
    """
    Puts the activity timestamps end into the given data.
    
    Parameters
    ----------
    end : `None`, `datetime`
        Activity timestamps end.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if defaults or (end is not None):
        if (end is not None):
            end = datetime_to_millisecond_unix_time(end)
        
        data['end'] = end
    
    return data


validate_end = nullable_date_time_validator_factory('end')

# start

def parse_start(data):
    """
    Parses activity timestamp time from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Activity timestamp data.
    
    Returns
    -------
    start : `None`, `datetime`
    """
    start = data.get('start', None)
    if (start is not None):
        return millisecond_unix_time_to_datetime(start)


def put_start(start, data, defaults):
    """
    Puts the activity timestamps start into the given data.
    
    Parameters
    ----------
    start : `None`, `datetime`
        Activity timestamps start.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if defaults or (start is not None):
        if (start is not None):
            start = datetime_to_millisecond_unix_time(start)
        
        data['start'] = start
    
    return data


validate_start = nullable_date_time_validator_factory('start')



__all__ = ()

from scarletio import Compound

from ...http import DiscordApiClient
from ...guild import BanAddMultipleResult, BanEntry
from ...guild.ban_add_multiple_result.fields import (
    put_delete_message_duration, put_user_ids, validate_delete_message_duration, validate_user_ids
)
from ...user import User
from ...utils import log_time_converter

from ..request_helpers import get_guild_id, get_user_id


def _assert__guild_ban_get_chunk__limit(limit):
    """
    Asserts the `limit` parameter of ``Client.guild_ban_get_chunk`` method.
    
    Parameters
    ----------
    limit : `int`
        The amount of ban entries to request. Can be in range [0:1000].
    
    Raises
    ------
    AssertionError
        - If `limit` was not given as `int`.
    """
    if (limit is not None) and (not isinstance(limit, int)):
        raise AssertionError(
            f'`limit` can be `None`, `int`, got {limit.__class__.__name__}; {limit!r}.'
        )
    
    return True


class ClientCompoundGuildBanEndpoints(Compound):
    
    api : DiscordApiClient
    
    
    async def guild_ban_add(self, guild, user, *, delete_message_duration = ..., reason = None):
        """
        Bans the given user from the guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild from where the user will be banned.
        
        user : ``ClientUserBase``, `int`
            The user to ban from the guild.
        
        delete_message_duration : `int`, Optional (Keyword only)
            How much seconds back the user's messages should be deleted. Can be in range [0:604800].
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        user_id = get_user_id(user)
        
        data = {}
        
        # delete_message_duration
        if (delete_message_duration is not ...):
            put_delete_message_duration(validate_delete_message_duration(delete_message_duration), data, False)
        
        await self.api.guild_ban_add(guild_id, user_id, data, reason)
    
    
    async def guild_ban_add_multiple(self, guild, user_ids, *, delete_message_duration = 0, reason = None):
        """
        Bans the given users from the guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild from where the user will be banned.
        
        user_ids : `None`, `iterable` of `int`, `iterable` of ``ClientUserBase``
            The user to ban from the guild.
        
        delete_message_duration : `int` = `0`, Optional (Keyword only)
            How much seconds back the user's messages should be deleted. Can be in range [0:604800].
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        data = {}
        
        # delete_message_duration
        if (delete_message_duration is not ...):
            put_delete_message_duration(validate_delete_message_duration(delete_message_duration), data, False)
        
        # user_ids
        put_user_ids(validate_user_ids(user_ids), data, False)
        
        data = await self.api.guild_ban_add_multiple(guild_id, data, reason)
        return BanAddMultipleResult.from_data(data)
    
    
    async def guild_ban_delete(self, guild, user, *, reason = None):
        """
        Unbans the user from the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild from where the user will be unbanned.
        user : ``ClientUserBase``, `int`
            The user to unban at the guild.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild`` nor `int`.
            - If `user` was not given neither as ``ClientUserBase``, nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        user_id = get_user_id(user)
        
        await self.api.guild_ban_delete(guild_id, user_id, reason)
    
    
    async def guild_ban_get_chunk(self, guild, *, after = None, before = None, limit = 0):
        """
        Returns a chunk of the guild's ban entries.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's bans will be requested
        after : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp after the banned users were created.
        before : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp before the banned users were created.
        limit : `int` = `0`, Optional (Keyword only)
            The amount of ban entries to request. Can be in range [0:1000]
            
            When given as non-positive (<= 0) defaults to the maximal amount.
        
        Returns
        -------
        ban_entries : `list` of ``BanEntry`` elements
            User, reason pairs for each ban entry.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild``, `int`.
            - If `after`, `before` was passed with an unexpected type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        See Also
        --------
        ``.guild_ban_get_all`` : Getting all ban entries of a guild.
        """
        guild_id = get_guild_id(guild)
        
        assert _assert__guild_ban_get_chunk__limit(limit)
        
        if limit <= 0 or limit > 1000:
            limit = 1000
        
        query_parameters = {}
        
        if limit != 1000:
            query_parameters['limit'] = limit
        
        if (after is not None):
            query_parameters['after'] = log_time_converter(after)
        
        if (before is not None):
            query_parameters['before'] = log_time_converter(before)
        
        ban_entry_datas = await self.api.guild_ban_get_chunk(guild_id, query_parameters)
        return [BanEntry.from_data(ban_entry_data) for  ban_entry_data in ban_entry_datas]
    
    
    async def guild_ban_get_all(self, guild):
        """
        Returns the guild's ban entries.
        
        This method might need multiple requests to complete it's task
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's bans will be requested
        
        Returns
        -------
        ban_entries : `list` of ``BanEntry`` elements
            User - reason pairs for each ban entry.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        See Also
        --------
        ``.guild_ban_get_chunk`` : Getting a chunk of ban entries up to 1000 users.
        """
        guild_id = get_guild_id(guild)
        

        query_parameters = {'after': 0}
        
        ban_entries = []
        
        while True:
            ban_entry_datas = await self.api.guild_ban_get_chunk(guild_id, query_parameters)
            ban_entries.extend(BanEntry.from_data(ban_entry_data) for  ban_entry_data in ban_entry_datas)
            
            if len(ban_entry_datas) < 1000:
                break
            
            query_parameters['after'] = ban_entries[-1].user.id
        
        return ban_entries
    
    
    async def guild_ban_get(self, guild, user):
        """
        Returns the guild's ban entry for the given user id.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild where the user banned.
        user : ``ClientUserBase``, `int`
            The user's or their id, who's entry is requested.
        
        Returns
        -------
        ban_entry : ``BanEntry``
            The ban entry.
        
        Raises
        ------
        TypeError
            - If `guild` was not passed neither as ``Guild``, `int`.
            - If `user` was not given neither as ``ClientUserBase``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        user_id = get_user_id(user)
        
        data = await self.api.guild_ban_get(guild_id, user_id)
        return BanEntry(User.from_data(data['user']), data.get('reason', None))



from .message_builder import *
from .message_reference_configuration import *
from .preinstanced import *


__all__ = (
    *message_builder.__all__,
    *message_reference_configuration.__all__,
    *preinstanced.__all__,
)



__all__ = ()

from scarletio import RichAttributeErrorBaseType

from .constants import (
    CONVERSION_KIND_FIELD, CONVERSION_KIND_INSTANCE, CONVERSION_KIND_KEYWORD, CONVERSION_KIND_NONE,
    CONVERSION_KIND_POSITIONAL
)


def _select_setter(kind, attribute_requester):
    """
    Selects the setter for the given conversion.
    
    Parameters
    ----------
    kind : `int`
        Conversion kind to select setter for.
    attribute_requester : `callable`
        Function allowing to request type attribute before the type is created.
    
    Returns
    -------
    setter : `FunctionType`
    
    Raises
    ------
    RuntimeError
        - Unexpected kind.
        - Missing attribute.
    """
    if kind == CONVERSION_KIND_NONE:
        setter_name = '_setter_none'
    elif kind == CONVERSION_KIND_FIELD:
        setter_name = '_setter_field'
    elif kind == CONVERSION_KIND_POSITIONAL:
        setter_name = '_setter_positional'
    elif kind == CONVERSION_KIND_KEYWORD:
        setter_name = '_setter_keyword'
    elif kind == CONVERSION_KIND_INSTANCE:
        setter_name = '_setter_instance'
    else:
        raise RuntimeError(f'Unexpected kind value. Got: {kind!r}.')
    
    return attribute_requester(setter_name)


def _select_getter(kind, attribute_requester):
    """
    Selects the getter for the given conversion.
    
    Parameters
    ----------
    kind : `int`
        Conversion kind to select setter for.
    attribute_requester : `callable`
        Function allowing to request type attribute before the type is created.
    
    Returns
    -------
    getter : `FunctionType`
    
    Raises
    ------
    RuntimeError
        - Unexpected kind.
        - Missing attribute.
    """
    if kind == CONVERSION_KIND_NONE:
        getter_name = '_getter_none'
    elif kind == CONVERSION_KIND_FIELD:
        getter_name = '_getter_field'
    elif kind == CONVERSION_KIND_POSITIONAL:
        getter_name = '_getter_none'
    elif kind == CONVERSION_KIND_KEYWORD:
        getter_name = '_getter_none'
    elif kind == CONVERSION_KIND_INSTANCE:
        getter_name = '_getter_none'
    else:
        raise RuntimeError(f'Unexpected kind value. Got: {kind!r}.')
    
    return attribute_requester(getter_name)


def _conversion_descriptor_sort_key(conversion_descriptor):
    """
    Sort key for conversions.
    
    Parameters
    ----------
    conversion_descriptor : ``ConversionDescriptor``
        Conversion to get its sort key of.
    
    Returns
    -------
    sort_key : `int`
    """
    return conversion_descriptor.conversion.sort_priority


class ConversionDescriptor(RichAttributeErrorBaseType):
    """
    Wraps a conversion into a descriptor linking it into the respective type.
    
    Attributes
    ----------
    attribute_name : `None | str`
        The attribute's name the descriptor represents.
    conversion : ``Conversion``
        The descripted conversion.
    getter : `FunctionType`
        Field getter to use on get operations.
    output_conversion : ``Conversion``
        Conversion that ``.conversion`` outputs. Can be the same.
    setter : `FunctionType`
        Field setter to use on set operations.
    """
    __slots__ = ('attribute_name', 'conversion', 'getter', 'output_conversion', 'setter')
    
    def __new__(cls, attribute_name, conversion, attribute_requester):
        """
        Creates a new conversion descriptor.
        
        Parameters
        ----------
        attribute_name : `None | str`
            The attribute's name the descriptor represents.
        conversion : ``Conversion``
            Conversion we are descripting. Is that even a word?
        attribute_requester : `callable`
            Function allowing to request type attribute before the type is created.
        """
        output_conversion = conversion.output_conversion
        if output_conversion is None:
            output_conversion = conversion
        
        getter = _select_getter(output_conversion.kind, attribute_requester)
        setter = _select_setter(output_conversion.kind, attribute_requester)
        
        self = object.__new__(cls)
        self.attribute_name = attribute_name
        self.conversion = conversion
        self.getter = getter
        self.output_conversion = output_conversion
        self.setter = setter
        return self
    
    
    def __set__(self, instance, value):
        """
        Sets the field with the defined conversion.
        
        Parameters
        ----------
        instance : ``BuilderBase``
            Instance to set the fields to.
        value : `object`
            Object to set.
        
        Raises
        ------
        RuntimeError
            - Field is not settable.
        TypeError
            - `value` of invalid type given.
        """
        conversion = self.conversion
        set_validator = conversion.set_validator
        if (set_validator is None):
            raise RuntimeError(
                f'Field {conversion.name!r} is not settable.'
            )
        
        for value in set_validator(value):
            self.setter(instance, self.output_conversion, value)
            return
        
        return self.raise_type_error(None, value)
    
    
    def __get__(self, instance, instance_type):
        """
        Gets the field with the conversion.
        If accessed as a type attribute returns itself.
        
        Parameters
        ----------
        instance : `None | BuilderBase`
            Instance to set the fields to.
        instance_type : ``BuilderMeta``
            The instance's type.
        
        Returns
        -------
        value : `object`
        
        Raises
        ------
        RuntimeError
            - Field is not gettable.
        """
        if instance is None:
            return self
        
        output_conversion = self.output_conversion
        value = self.getter(instance, output_conversion)
        
        conversion = self.conversion
        if conversion is not output_conversion:
            get_processor = conversion.get_processor
            if (get_processor is not None):
                value = get_processor(value)
        
        return value
    
    
    def raise_type_error(self, key, value):
        """
        Raises context fit `TypeError` with they given `key` and `value`.
        This function always raises. Use `return ...` to help you IDE identify that it is a function exit.
        
        Parameters
        ----------
        key : `None | str`
            Conversion name to use. Can be passed as `None` to default.
        value : `object`
            The value that should be raised exception with.
        
        Raises
        ------
        TypeError
        """
        conversion = self.conversion
        if key is None:
            key = self.attribute_name
            if key is None:
                key = conversion.name
        
        raise TypeError(
            f'{key!s} can be {conversion.expected_types_messages!s}. Got {type(value).__name__}; {value!r}.'
        )
    
    
    def __repr__(self):
        """Returns the descriptor's representation."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' conversion = ')
        repr_parts.append(repr(self.conversion))
        
        attribute_name = self.attribute_name
        if (attribute_name is not None):
            repr_parts.append(' attribute_name = ')
            repr_parts.append(repr(attribute_name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)



__all__ = ()

from scarletio import RichAttributeErrorBaseType

from ...discord.allowed_mentions import AllowedMentionProxy
from ...discord.message import MessageFlag
from ...discord.message.message_builder.conversions import CONVERSION_ALLOWED_MENTIONS, CONVERSION_FLAGS
from ...discord.preconverters import preconvert_bool


MESSAGE_FLAG_VALUE_SHOW_FOR_INVOKING_USER_ONLY = MessageFlag().update_by_keys(invoking_user_only = True)


def _validate_show_for_invoking_user_only(show_for_invoking_user_only):
    """
    Validates the given `show_for_invoking_user_only` value.
    
    Parameters
    ----------
    show_for_invoking_user_only : `bool`
        The `show_for_invoking_user_only` value to validate.
    
    Returns
    -------
    show_for_invoking_user_only : `None` or `bool`
        The validated `show_for_invoking_user_only` value.
    
    Raises
    ------
    TypeError
        If `show_for_invoking_user_only` was not given as `bool`.
    """
    show_for_invoking_user_only = preconvert_bool(show_for_invoking_user_only, 'show_for_invoking_user_only')
    
    return show_for_invoking_user_only


def _validate_allowed_mentions(allowed_mentions):
    """
    Validates the given `allowed_mentions` parameter.
    
    Parameters
    ----------
    allowed_mentions : `None`, `str`, ``UserBase``, ``Role``, ``AllowedMentionProxy``, \
            `list` of (`str`, ``UserBase``, ``Role`` )
         Which user or role can the response message ping (or everyone).
    
    Returns
    -------
    allowed_mention_proxy : ``AllowedMentionProxy``
    """
    if allowed_mentions is None:
        allowed_mention_proxy = AllowedMentionProxy()
    elif isinstance(allowed_mentions, AllowedMentionProxy):
        allowed_mention_proxy = allowed_mentions.copy()
    elif isinstance(allowed_mentions, list):
        allowed_mention_proxy = AllowedMentionProxy(*allowed_mentions)
    else:
        allowed_mention_proxy = AllowedMentionProxy(allowed_mentions)
    
    return allowed_mention_proxy


def _validate_wait_for_acknowledgement(wait_for_acknowledgement):
    """
    Validates the given `wait_for_acknowledgement` value.
    
    Parameters
    ----------
    wait_for_acknowledgement : `bool`
        The `wait_for_acknowledgement` value to validate.
    
    Returns
    -------
    wait_for_acknowledgement : `None` or `bool`
        The validated `wait_for_acknowledgement` value.
    
    Raises
    ------
    TypeError
        If `wait_for_acknowledgement` was not given as `bool`.
    """
    wait_for_acknowledgement = preconvert_bool(wait_for_acknowledgement, 'wait_for_acknowledgement')
    
    return wait_for_acknowledgement


class ResponseModifier(RichAttributeErrorBaseType):
    """
    Modifies values returned and yielded to command coroutine processor.
    
    Attributes
    ----------
    allowed_mentions : `None`, ``AllowedMentionProxy``
         Which user or role can the response message ping (or everyone).
    wait_for_acknowledgement : `bool`
        Whether acknowledge tasks should be ensure asynchronously.
    show_for_invoking_user_only : `None`, `bool`
        Whether the response message should only be shown for the invoking user.
    """
    __slots__ = ('allowed_mentions', 'wait_for_acknowledgement', 'show_for_invoking_user_only',)
    
    def __new__(cls, keyword_parameters):
        """
        Creates a new request modifiers from additionally defined
        
        Parameters
        ----------
        keyword_parameters : `dict<str, object>`
            Additional keyword parameters.
        
        Returns
        -------
        self : `None` / ``ResponseModifier``
            Returns `None` if there are no request modifiers specified.
        
        Raises
        ------
        TypeError
            If a parameter's type is incorrect.
        ValueError
            If a parameter's value is incorrect.
        """
        if keyword_parameters:
            parameters_found = False
            
            try:
                allowed_mentions = keyword_parameters.pop('allowed_mentions')
            except KeyError:
                allowed_mentions = None
            else:
                allowed_mentions = _validate_allowed_mentions(allowed_mentions)
                parameters_found = True
            
            try:
                show_for_invoking_user_only = keyword_parameters.pop('show_for_invoking_user_only')
            except KeyError:
                show_for_invoking_user_only = None
            else:
                show_for_invoking_user_only = _validate_show_for_invoking_user_only(show_for_invoking_user_only)
                parameters_found = True
            
            try:
                wait_for_acknowledgement = keyword_parameters.pop('wait_for_acknowledgement')
            except KeyError:
                wait_for_acknowledgement = None
            else:
                wait_for_acknowledgement = _validate_wait_for_acknowledgement(wait_for_acknowledgement)
                parameters_found = True
            
            if parameters_found:
                self = object.__new__(cls)
                self.allowed_mentions = allowed_mentions
                self.wait_for_acknowledgement = wait_for_acknowledgement
                self.show_for_invoking_user_only = show_for_invoking_user_only
            else:
                self = None
        else:
            self = None
        
        return self
    
    
    def apply_to_creation(self, interaction_response):
        """
        Applies the response creation modifiers to the given parameters.
        
        Parameters
        ----------
        interaction_response : ``InteractionResponse``
            Interaction response to apply self to.
        """
        self._apply_to_shared(interaction_response)
        
        show_for_invoking_user_only = self.show_for_invoking_user_only
        if (show_for_invoking_user_only is not None):
            # This is a nested + nested field, so passing it as False does nothing.
            interaction_response._setter_field(
                CONVERSION_FLAGS,
                (MESSAGE_FLAG_VALUE_SHOW_FOR_INVOKING_USER_ONLY if show_for_invoking_user_only else False),
            )
    
    
    def apply_to_edition(self, interaction_response):
        """
        Applies the response edition modifiers to the given parameters.
        
        Parameters
        ----------
        interaction_response : ``InteractionResponse``
            Interaction response to apply self to.
        """
        self._apply_to_shared(interaction_response)
    
    
    def _apply_to_shared(self, interaction_response):
        """
        Applies creation and edition modifiers to the given parameters.
        
        Parameters
        ----------
        interaction_response : ``InteractionResponse``
            Interaction response to apply self to.
        """
        allowed_mentions = self.allowed_mentions
        if (allowed_mentions is not None):
            for _ in interaction_response._try_pull_field_value(CONVERSION_ALLOWED_MENTIONS):
                break
            else:
                interaction_response._setter_field(
                    CONVERSION_ALLOWED_MENTIONS,
                    allowed_mentions,
                )
    
    
    def __repr__(self):
        """Returns the parameter modifier's representation."""
        repr_parts = ['<', type(self).__name__]
        
        
        allowed_mentions = self.allowed_mentions
        if (allowed_mentions is not None):
            repr_parts.append(' allowed_mentions = ')
            repr_parts.append(repr(allowed_mentions))
            
            field_added = True
        
        else:
            field_added = False
        
        
        show_for_invoking_user_only = self.show_for_invoking_user_only
        if (show_for_invoking_user_only is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' show_for_invoking_user_only = ')
            repr_parts.append(repr(show_for_invoking_user_only))
        
        

        wait_for_acknowledgement = self.wait_for_acknowledgement
        if (wait_for_acknowledgement is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' wait_for_acknowledgement = ')
            repr_parts.append(repr(wait_for_acknowledgement))
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the parameter modifier's hash value."""
        hash_value = 0
        
        allowed_mentions = self.allowed_mentions
        if (allowed_mentions is not None):
            hash_value ^= hash(allowed_mentions)
        
        show_for_invoking_user_only = self.show_for_invoking_user_only
        if (show_for_invoking_user_only is not None):
            hash_value ^= show_for_invoking_user_only

        wait_for_acknowledgement = self.wait_for_acknowledgement
        if (wait_for_acknowledgement is not None):
            hash_value ^= wait_for_acknowledgement << 4
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two parameter modifiers are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.allowed_mentions != other.allowed_mentions:
            return False
        
        if self.wait_for_acknowledgement != other.wait_for_acknowledgement:
            return False
        
        if self.show_for_invoking_user_only != other.show_for_invoking_user_only:
            return False
        
        return True


def get_show_for_invoking_user_only_of(response_modifier):
    """
    Gets the `show_for_invoking_user_only` value of the given response modifier.
    
    Parameters
    ----------
    response_modifier : `None`, ``ResponseModifier``
        The respective response modifier if any,
    
    Returns
    -------
    show_for_invoking_user_only : `bool`
    """
    if response_modifier is None:
        show_for_invoking_user_only = False
    else:
        show_for_invoking_user_only = response_modifier.show_for_invoking_user_only
        if show_for_invoking_user_only is None:
            show_for_invoking_user_only = False
    
    return show_for_invoking_user_only


def get_show_for_invoking_user_only_from(interaction_response, response_modifier):
    """
    Gets the `show_for_invoking_user_only` value from the given parameters from the given response modifier if not
    present.
    
    Parameters
    ----------
    interaction_response : ``InteractionResponse``
        Interaction response to pull from.
    response_modifier : `None`, ``ResponseModifier``
        The respective response modifier if any,
    
    Returns
    -------
    show_for_invoking_user_only : `bool`
    """
    for value in interaction_response._try_pull_field_value(CONVERSION_FLAGS):
        return (True if value & MESSAGE_FLAG_VALUE_SHOW_FOR_INVOKING_USER_ONLY else False)
    
    return  get_show_for_invoking_user_only_of(response_modifier)


def get_wait_for_acknowledgement_of(response_modifier):
    """
    Gets the `wait_for_acknowledgement` value of the given response modifier.
    
    Parameters
    ----------
    wait_for_acknowledgement : `None`, ``ResponseModifier``
        The respective response modifier if any,
    
    Returns
    -------
    wait_for_acknowledgement : `bool`
    """
    if response_modifier is None:
        wait_for_acknowledgement = False
    else:
        wait_for_acknowledgement = response_modifier.wait_for_acknowledgement
    
    return wait_for_acknowledgement



__all__ = ('ApplicationCommandOptionMetadataInteger',)

from .numeric import ApplicationCommandOptionMetadataNumeric


class ApplicationCommandOptionMetadataInteger(ApplicationCommandOptionMetadataNumeric):
    """
    Integer parameter application command option metadata.
    
    Attributes
    ----------
    autocomplete : `bool`
        Whether the option supports auto completion.
    
    choices : `None`, `tuple` of ``ApplicationCommandOptionChoice``
        Choices for the user to pick from.
    
    max_value : `None`, `int`
        The maximal value permitted for this option.
    
    min_value : `None`, `int`
        The minimum value permitted for this option.
    
    required : `bool`
        Whether the parameter is required. Defaults to `False`.
    """
    TYPE = int
    
    __slots__ = ()



__all__ = ()


USER_MATCH_WEIGHT_NAME = 1
USER_MATCH_WEIGHT_DISPLAY_NAME = 2
USER_MATCH_WEIGHT_NICK = 3


def _parse_name_with_discriminator(name):
    """
    Tries to parse the user's name and their discriminator from the given name value.
    
    Parameters
    ----------
    name : `str`
        The name value to parse.
    
    Returns
    -------
    user_with_discriminator : `None`, `tuple` (`str`, `int`)
        Returns `None` on unsuccessful parsing.
    """
    separator_index = name.find('#')
    if separator_index < 2:
        return None
    
    if len(name) - separator_index != 5:
        return None
    
    try:
        discriminator = int(name[separator_index + 1 :])
    except ValueError:
        return None
    
    return name[: separator_index], discriminator
  
  
def _is_user_matching_name_with_discriminator(user, name_with_discriminator):
    """
    Returns whether the user matches the given name - discriminator pair.
    
    Parameters
    ----------
    user : ``ClientUserBase``
        The user we are trying to match.
    name_with_discriminator : `tuple` (`str`, `int`)
        Name - discriminator pair.
    
    Returns
    -------
    is_matching : `bool`
    """
    if user.discriminator != name_with_discriminator[1]:
        return False
   
    if user.name != name_with_discriminator[0]:
        return False
    
    return True


def _user_date_sort_key(item):
    """
    Sort key used inside ``Guild.get_users_like_ordered`` and in ``Guild.boosters`` to sort users by a specified date.
    
    Parameters
    ----------
    item : `tuple` (``ClientUserBase``, `datetime`)
        The user and it's specific date.
    
    Returns
    -------
    date : `datetime`
    """
    return item[1]


def _user_match_sort_key(item):
    """
    Sort key used inside of ``Guild.get_users_like`` to sort users based on their match rate.
    
    Parameters
    ----------
    item : `tuple` (``ClientUserBase``, `tuple` (`int`, `int`, `int`))
        The user and it's match rate.
    
    Returns
    -------
    match_rate : `tuple` (`int`, `int`, `int`)
    """
    return item[1]



__all__ = ('Stage',)

from scarletio import export

from ...bases import DiscordEntity
from ...channel import ChannelType, create_partial_channel_from_id
from ...core import GUILDS, STAGES
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...scheduled_event import PrivacyLevel

from .fields import (
    parse_channel_id, parse_discoverable, parse_guild_id, parse_id, parse_invite_code, parse_privacy_level,
    parse_scheduled_event_id, parse_topic, put_channel_id, put_discoverable, put_guild_id, put_id,
    put_invite_code, put_privacy_level, put_scheduled_event_id, put_topic, validate_channel_id,
    validate_discoverable, validate_guild_id, validate_id, validate_invite_code, validate_privacy_level,
    validate_scheduled_event_id, validate_topic
)


PRECREATE_FIELDS = {
    'channel': ('channel_id', validate_channel_id),
    'channel_id': ('channel_id', validate_channel_id),
    'discoverable': ('discoverable', validate_discoverable),
    'guild': ('guild_id', validate_guild_id),
    'guild_id': ('guild_id', validate_guild_id),
    'invite_code': ('invite_code', validate_invite_code),
    'privacy_level': ('privacy_level', validate_privacy_level),
    'scheduled_event': ('scheduled_event_id', validate_scheduled_event_id),
    'scheduled_event_id': ('scheduled_event_id', validate_scheduled_event_id),
    'topic': ('topic', validate_topic),
}


@export
class Stage(DiscordEntity, immortal = True):
    """
    Represents an active stage instance of a stage channel.
    
    Attributes
    ----------
    channel_id : `int`
        The stage channel's identifier where the stage is active.
    discoverable : `bool`
        Whether the stage is discoverable. Only applies for public stages.
    guild_id : `int`
        The stage guild's identifier.
    id : `int`
        The stage instance's identifier.
    invite_code : `None`, `str`
        Invite code to the stage's channel.
    privacy_level : ``PrivacyLevel``
        The privacy level of the stage.
    scheduled_event_id : `int`
        The scheduled event's identifier that started the stage.
    topic : `None`, `str`
        The topic of the stage. Can be empty string.
    """
    __slots__ = (
        'channel_id', 'discoverable', 'guild_id', 'invite_code', 'privacy_level', 'scheduled_event_id', 'topic'
    )
    
    
    def __new__(cls, *, privacy_level = ..., scheduled_event_id = ..., topic = ...):
        """
        Creates a partial stage with the given fields.
        
        Parameters
        ----------
        privacy_level : ``PrivacyLevel``, `int`, Optional (Keyword only)
            The privacy level of the stage.
        
        scheduled_event_id : `int`, ``ScheduledEvent``, Optional (Keyword only)
            The scheduled event's identifier that started the stage.
        
        topic : `None`, `str`, Optional (Keyword only)
            The topic of the stage. Can be empty string.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # privacy_level
        if privacy_level is ...:
            privacy_level = PrivacyLevel.guild_only
        else:
            privacy_level = validate_privacy_level(privacy_level)
        
        # scheduled_event_id
        if scheduled_event_id is ...:
            scheduled_event_id = 0
        else:
            scheduled_event_id = validate_scheduled_event_id(scheduled_event_id)
        
        # topic
        if topic is ...:
            topic = None
        else:
            topic = validate_topic(topic)
        
        # Construct
        self = object.__new__(cls)
        self.channel_id = 0
        self.discoverable = True
        self.guild_id = 0
        self.id = 0
        self.invite_code = None
        self.privacy_level = privacy_level
        self.scheduled_event_id = scheduled_event_id
        self.topic = topic
        return self
    
    
    @classmethod
    def from_data(cls, data, *, strong_cache = True):
        """
        Creates a new stage instance from the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Stage data.
        strong_cache : `bool` = `True`, Optional (Keyword only)
            Whether the instance should be put into its strong cache.
        
        Returns
        -------
        self : `instance<cls>`
        """
        stage_id = parse_id(data)
        try:
            self = STAGES[stage_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = stage_id
            self._set_attributes(data)
            STAGES[stage_id] = self
        
        else:
            if strong_cache and (not self.partial):
                return self
            
            self._set_attributes(data)
        
        if strong_cache:
            try:
                guild = GUILDS[self.guild_id]
            except KeyError:
                pass
            else:
                stages = guild.stages
                if stages is None:
                    stages = guild.stages = {}
                
                stages[stage_id] = self
        
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Serialises the stage to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether we want to include identifiers as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) Items
        """
        data = {}
        put_privacy_level(self.privacy_level, data, defaults)
        put_scheduled_event_id(self.scheduled_event_id, data, defaults)
        put_topic(self.topic, data, defaults)
        
        if include_internals:
            put_channel_id(self.channel_id, data, defaults)
            put_discoverable(self.discoverable, data, defaults)
            put_guild_id(self.guild_id, data, defaults)
            put_id(self.id, data, defaults)
            put_invite_code(self.invite_code, data, defaults)
        
        return data
    
    
    def _set_attributes(self, data):
        """
        Sets the stage's attributes from the given data. Excludes `.id`.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Stage data.
        """
        self.channel_id = parse_channel_id(data)
        self.guild_id = parse_guild_id(data)
        self.scheduled_event_id = parse_scheduled_event_id(data)
    
        self._update_attributes(data)
    
    
    def _update_attributes(self, data):
        """
        Updates the stage from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Stage data.
        """
        self.discoverable = parse_discoverable(data)
        self.invite_code = parse_invite_code(data)
        self.privacy_level = parse_privacy_level(data)
        self.topic = parse_topic(data)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the stage from the given data and returns the changed attributes in `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Stage data.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `object`) items
            The changed attributes of the stage.
            
            Each item in the returned dictionary is optional.
        
        Returned Data Structure
        -----------------------
        
        +---------------+-----------------------+
        | Keys          | Values                |
        +===============+=======================+
        | discoverable  | `bool`                |
        +---------------+-----------------------+
        | invite_code   | `None`, `str`         |
        +---------------+-----------------------+
        | privacy_level | ``PrivacyLevel``      |
        +---------------+-----------------------+
        | topic         | `None`, `str`         |
        +---------------+-----------------------+
        """
        old_attributes = {}
        
        discoverable = parse_discoverable(data)
        if discoverable != self.discoverable:
            old_attributes['discoverable'] = self.discoverable
            self.discoverable = discoverable
        
        invite_code = parse_invite_code(data)
        if invite_code != self.invite_code:
            old_attributes['invite_code'] = self.invite_code
            self.invite_code = invite_code
        
        privacy_level = parse_privacy_level(data)
        if privacy_level is not self.privacy_level:
            old_attributes['privacy_level'] = self.privacy_level
            self.privacy_level = privacy_level
        
        topic = parse_topic(data)
        if topic != self.topic:
            old_attributes['topic'] = self.topic
            self.topic = topic
        
        return old_attributes
    
    
    def __repr__(self):
        """Returns the stage's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        repr_parts.append(' guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        repr_parts.append(', channel_id = ')
        repr_parts.append(repr(self.channel_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Hashes the stage."""
        stage_id = self.id
        if stage_id:
            return stage_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Returns the stage's hash value.
        This function is called when the stage is partial.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        # channel_id | internal
        # discoverable | internal
        # guild_id | internal
        # id | internal
        # invite_code | internal
        hash_value ^= hash(self.privacy_level)
        
        hash_value ^= self.scheduled_event_id
        
        topic = self.topic
        if (topic is not None):
            hash_value ^= hash(topic)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two stages are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two stages are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self is equal to other. Other must be same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self.id != other.id:
                return False
        
        # privacy_level
        if self.privacy_level is not other.privacy_level:
            return False
        
        # scheduled_event_id
        if self.scheduled_event_id != other.scheduled_event_id:
            return False
        
        # topic
        if self.topic != other.topic:
            return False
        
        return True
    
    
    @classmethod
    def _create_empty(cls, stage_id):
        """
        Creates a stage instance with its attributes set to their default values.
        
        Parameters
        ----------
        stage_id : `int`
            The stage's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.channel_id = 0
        self.discoverable = True
        self.guild_id = 0
        self.id = stage_id
        self.invite_code = None
        self.privacy_level = PrivacyLevel.guild_only
        self.scheduled_event_id = 0
        self.topic = None
        return self
    
    
    @classmethod
    def precreate(cls, stage_id, **keyword_parameters):
        """
        Precreates the stage by creating a partial one with the given parameters. When the stage is loaded
        the precreated one will be picked up. If an already existing stage would be precreated, returns that
        instead and updates that only, if that is partial.
        
        Parameters
        ----------
        stage_id : `int`
            The stage's identifier.
        **keyword_parameters : keyword parameters
            Additional predefined attributes for the stage.
        
        Other Parameters
        ----------------
        channel : `int`, ``Channel``, Optional (Keyword only)
            Alternative for `channel_id`.
        channel_id : `int`, ``Channel``, Optional (Keyword only)
            The stage channel or its identifier where the stage is active.
        discoverable : `bool`, Optional (Keyword only)
            Whether the stage is discoverable. Only applies for public stages.
        guild : `int`, ``Guild``, Optional (Keyword only)
            Alternative for `guild_id`.
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The stage's guild or its identifier.
        invite_code : `None`, `str`, Optional (Keyword only)
            Invite code to the stage's channel.
        privacy_level : ``PrivacyLevel``, `int`, Optional (Keyword only)
            The privacy level of the stage.
        scheduled_event : `int`, ``ScheduledEvent``, Optional (Keyword only)
            Alternative for `scheduled_event_id`.
        scheduled_event_id : `int`, ``ScheduledEvent``, Optional (Keyword only)
            The scheduled event or its identifier that started the stage.
        topic : `None`, `str`, Optional (Keyword only)
            The topic of the stage. Can be empty string.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        stage_id = validate_id(stage_id)

        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = STAGES[stage_id]
        except KeyError:
            self = cls._create_empty(stage_id)
            STAGES[stage_id] = self
        else:
            if not self.partial:
                return self
        
        if (processed is not None):
            for name, value in processed:
                setattr(self, name, value)
        
        return self
    
    
    def copy(self):
        """
        Copies the stage state.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.channel_id = 0
        new.discoverable = True
        new.guild_id = 0
        new.id = 0
        new.invite_code = None
        new.privacy_level = self.privacy_level
        new.scheduled_event_id = self.scheduled_event_id
        new.topic = self.topic
        return new
    
    
    def copy_with(self, *, privacy_level = ..., scheduled_event_id = ..., topic = ...):
        """
        Copies the stage with the given fields.
        
        Parameters
        ----------
        privacy_level : ``PrivacyLevel``, `int`, Optional (Keyword only)
            The privacy level of the stage.
        
        scheduled_event_id : `int`, ``ScheduledEvent``, Optional (Keyword only)
            The scheduled event's identifier that started the stage.
        
        topic : `None`, `str`, Optional (Keyword only)
            The topic of the stage. Can be empty string.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's value is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # privacy_level
        if privacy_level is ...:
            privacy_level = self.privacy_level
        else:
            privacy_level = validate_privacy_level(privacy_level)
        
        # scheduled_event_id
        if scheduled_event_id is ...:
            scheduled_event_id = self.scheduled_event_id
        else:
            scheduled_event_id = validate_scheduled_event_id(scheduled_event_id)
        
        # topic
        if topic is ...:
            topic = self.topic
        else:
            topic = validate_topic(topic)
        
        # Construct
        new = object.__new__(type(self))
        new.channel_id = 0
        new.discoverable = True
        new.guild_id = 0
        new.id = 0
        new.invite_code = None
        new.privacy_level = privacy_level
        new.scheduled_event_id = scheduled_event_id
        new.topic = topic
        return new
    
    
    def _delete(self):
        """
        Removes the stage's references.
        """
        try:
            guild = GUILDS[self.guild_id]
        except KeyError:
            pass
        else:
            stages = guild.stages
            if (stages is not None):
                try:
                    del stages[self.id]
                except KeyError:
                    pass
                else:
                    if not stages:
                        guild.stages = None
    
    
    @property
    def partial(self):
        """
        Returns whether the stage is partial.
        
        Returns
        -------
        partial : `bool`
        """
        guild_id = self.guild_id
        if guild_id:
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                pass
            else:
                if not guild.partial:
                    stages = guild.stages
                    if (stages is not None) and (self.id in stages.keys()):
                        return False
        
        return True
    
    
    @property
    def channel(self):
        """
        Returns the stage's channel.
        
        Returns
        -------
        channel : `None`, ``Channel``
        """
        return create_partial_channel_from_id(self.channel_id, ChannelType.guild_stage, self.guild_id)
    
    
    @property
    def guild(self):
        """
        Returns the stage's guild.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)



from .application_command import *
from .auto_moderation_action_execution import *
from .channel import *
from .channel_permission_overwrite import *
from .guild import *
from .message import *
from .stage import *
from .user import *


__all__ = (
    *application_command.__all__,
    *auto_moderation_action_execution.__all__,
    *channel.__all__,
    *channel_permission_overwrite.__all__,
    *guild.__all__,
    *message.__all__,
    *stage.__all__,
    *user.__all__,
)



from .utils import *

from .compatibility import *
from .constants import *
from .helpers import *
from .plugin_auto_reloader import *


__all__ = (
    *utils.__all__,
    *compatibility.__all__,
    *constants.__all__,
    *helpers.__all__,
    *plugin_auto_reloader.__all__,
)


from .. import register_library_extension
register_library_extension('HuyaneMatsu.plugin_auto_reloader')
del register_library_extension



__all__ = ()

from ...field_parsers import nullable_string_parser_factory
from ...field_putters import nullable_string_optional_putter_factory
from ...field_validators import int_conditional_validator_factory, nullable_string_validator_factory

# id

parse_id = nullable_string_parser_factory('id')
put_id = nullable_string_optional_putter_factory('id')
validate_id = nullable_string_validator_factory('party_id', 0, 1024)

# size # max

def parse_size_and_max(data):
    """
    Parses out the (current) size and the max (size) fields from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Activity party data.
    
    Returns
    -------
    size : `int`
    max_ : `int`
    """
    return data.get('size', (0, 0))


def put_size_and_max(size_and_max, data, defaults):
    """
    Puts the given size and max value pair into the given data.
    
    Parameters
    ----------
    size_and_max : `tuple` (`int`, `int`)
        Size and max pair.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if defaults or any(size_and_max):
        data['size'] = size_and_max
    return data

# size

validate_size = int_conditional_validator_factory(
    'size',
    0,
    lambda size : size >= 0,
    '>= 0',
)

# max

validate_max = int_conditional_validator_factory(
    'max',
    0,
    lambda max_ : max_ >= 0,
    '>= 0',
)



__all__ = ()

from ...application import Application, Entitlement
from ...channel import Channel, ChannelType, create_partial_channel_from_id
from ...field_parsers import (
    default_entity_parser_factory, entity_id_parser_factory, flag_parser_factory, force_string_parser_factory,
    nullable_entity_array_parser_factory, nullable_entity_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    entity_id_optional_putter_factory, entity_id_putter_factory, entity_putter_factory, force_string_putter_factory,
    nullable_entity_array_optional_putter_factory, preinstanced_putter_factory, string_flag_putter_factory
)
from ...field_validators import (
    default_entity_validator_factory, entity_id_validator_factory, flag_validator_factory,
    force_string_validator_factory, nullable_entity_array_validator_factory, nullable_entity_validator_factory,
    preinstanced_validator_factory
)
from ...guild import (
    Guild, create_interaction_guild_data, create_partial_guild_from_id, create_partial_guild_from_interaction_guild_data
)
from ...localization import Locale
from ...localization.utils import LOCALE_DEFAULT
from ...message import Message
from ...message.message_interaction.fields import (
    parse_authorizer_user_ids, put_authorizer_user_ids, validate_authorizer_user_ids
)
from ...permission import Permission
from ...permission.permission import PERMISSION_PRIVATE
from ...user import ClientUserBase, User, ZEROUSER

from ..interaction_metadata import InteractionMetadataBase

from .preinstanced import InteractionType

# application_id

parse_application_id = entity_id_parser_factory('application_id')
put_application_id = entity_id_optional_putter_factory('application_id')
validate_application_id = entity_id_validator_factory('application_id', Application)

# application_permissions

parse_application_permissions = flag_parser_factory('app_permissions', Permission)
put_application_permissions = string_flag_putter_factory('app_permissions')
validate_application_permissions = flag_validator_factory('application_permissions', Permission)

# authorizer_user_ids

# Imported from `...message.message_interaction.fields`.
# They are the same.
# It does not matter where we define them.
#
# parse_authorizer_user_ids = ...
# put_authorizer_user_ids = ...
# validate_authorizer_user_ids = ...

# channel

parse_channel = default_entity_parser_factory(
    'channel', Channel, default_factory = lambda : create_partial_channel_from_id(0, ChannelType.unknown, 0)
)
put_channel = entity_putter_factory('channel', Channel, force_include_internals = True)
validate_channel = default_entity_validator_factory(
    'channel', Channel, default_factory = lambda : create_partial_channel_from_id(0, ChannelType.unknown, 0)
)

# entitlements

parse_entitlements = nullable_entity_array_parser_factory('entitlements', Entitlement)
put_entitlements = nullable_entity_array_optional_putter_factory(
    'entitlements', Entitlement, force_include_internals = True,
)
validate_entitlements = nullable_entity_array_validator_factory('entitlements', Entitlement)

# guild

def parse_guild(data):
    """
    Parses out the interaction's guild from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    guild : `None | Guild`
    """
    guild_data = data.get('guild', None)
    if (guild_data is not None):
        return create_partial_guild_from_interaction_guild_data(guild_data)
    
    guild_id = data.get('guild_id', None)
    if (guild_id is not None):
        return create_partial_guild_from_id(int(guild_id))
    
    return None


def put_guild(guild, data, defaults):
    """
    Puts the given `guild''s data into the given interaction data.
    
    Parameters
    ----------
    guild : `None | Guild`
        The guild to serialize.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default fields should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if (guild is not None) or defaults:
        if guild is None:
            guild_data = None
            guild_id = None
            guild_locale = None
        else:
            guild_data = create_interaction_guild_data(guild)
            guild_id = str(guild.id)
            guild_locale = guild.locale.value
        
        data['guild'] = guild_data
        data['guild_id'] = guild_id
        data['guild_locale'] = guild_locale
    
    return data


validate_guild = nullable_entity_validator_factory('guild', Guild)

# guild_locale

parse_guild_locale = preinstanced_parser_factory('guild_locale', Locale, LOCALE_DEFAULT)
put_guild_locale = preinstanced_putter_factory('guild_locale')
validate_guild_locale = preinstanced_validator_factory('guild_locale', Locale)

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('id')

# interaction

def validate_interaction(interaction, interaction_type):
    """
    Validates the given `interaction`.
    
    Parameters
    ----------
    interaction : `None`, ``InteractionMetadataBase``
        Interaction metadata.
    
    interaction_type : ``InteractionType``
        The respective interaction type.
    
    Returns
    -------
    interaction : ``InteractionMetadataBase``
    
    Raises
    ------
    TypeError
        - If `interaction`-s type is incorrect.
    """
    if interaction is None:
        return interaction_type.metadata_type()
    
    if not isinstance(interaction, InteractionMetadataBase):
        raise TypeError(
            f'`interaction` can be `None`, `{InteractionMetadataBase.__name__}` instance, got '
            f'{interaction.__class__.__name__}; {interaction!r}.'
        )
    
    if not isinstance(interaction, interaction_type.metadata_type):
        raise TypeError(
            f'Interactions of `{interaction_type!r}` type event can only be '
            f'`{interaction_type.metadata_type.__name__}` instances. '
            f'Make sure to pass the correct `interaction_type` parameter. '
            f'Got {interaction.__class__.__name__}; {interaction!r}.'
        )
    
    return interaction

# message

parse_message = nullable_entity_parser_factory('message', Message)

def put_message(message, data, defaults):
    """
    Puts the given `message` into the given data.
    
    Parameters
    ----------
    message : `None`, ``Message``
        The message to put into the given `data`.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default fields should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if defaults or (message is not None):
        if message is None:
            entity_data = None
        else:
            entity_data = message.to_data(defaults = defaults, include_internals = True)
        
        data['message'] = entity_data
    
    return data

validate_message = nullable_entity_validator_factory('message', Message)

# token

parse_token = force_string_parser_factory('token')
put_token = force_string_putter_factory('token')
validate_token = force_string_validator_factory('token', 0, 1024)

# type

parse_type = preinstanced_parser_factory('type', InteractionType, InteractionType.none)
put_type = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('interaction_type', InteractionType)

# user & user_permissions

def parse_user(data, guild_id = 0):
    """
    Parses out the interaction event's user from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Interaction event data.
    guild_id : `int` = `0`, Optional (Keyword only)
        The respective guild's identifier.
    
    Returns
    -------
    user : ``ClientUserBase``
    """
    # user
    try:
        guild_profile_data = data['member']
    except KeyError:
        try:
            user_data = data['user']
        except KeyError:
            return ZEROUSER
        
        guild_profile_data = None
    else:
        try:
            user_data = guild_profile_data['user']
        except KeyError:
            return ZEROUSER
    
    return User.from_data(user_data, guild_profile_data, guild_id)


def put_user(user, data, defaults, *, guild_id = 0):
    """
    Puts the given `user` into the given data.
    
    Parameters
    ----------
    user : ``ClientUserBase``
        The user to put into the given `data`.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default fields should be included as well.
    guild_id : `int` = `0`, Optional (Keyword only)
        The user's specific guild's identifier to use for getting the user's local profile.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    try:
        guild_profile = user.guild_profiles[guild_id]
    except KeyError:
        data['user'] = user.to_data(defaults = defaults, include_internals = True)
    
    else:
        guild_profile_data = guild_profile.to_data(defaults = defaults, include_internals = True)
        guild_profile_data['user'] = user.to_data(defaults = defaults, include_internals = True)
        
        data['member'] = guild_profile_data
    
    return data


validate_user = default_entity_validator_factory('user', ClientUserBase, default = ZEROUSER)

# user_locale

parse_user_locale = preinstanced_parser_factory('locale', Locale, LOCALE_DEFAULT)
put_user_locale = preinstanced_putter_factory('locale')
validate_user_locale = preinstanced_validator_factory('user_locale', Locale)

# user_permissions

def parse_user_permissions(data):
    """
    Parses out the interaction's user's permissions from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Interaction data.
    
    Return
    ------
    user_permissions : ``Permission``
    """
    try:
        guild_profile_data = data['member']
    except KeyError:
        return PERMISSION_PRIVATE
    
    try:
        user_permissions = guild_profile_data['permissions']
    except KeyError:
        return PERMISSION_PRIVATE
    
    return Permission(user_permissions)


def put_user_permissions(user_permissions, data, defaults):
    """
    Puts the given `user_permissions` into the given data.
    
    Parameters
    ----------
    user_permissions : ``Permission``
        The permission to put into the given `data`.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default fields should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    try:
        guild_profile_data = data['member']
    except KeyError:
        pass
    else:
        guild_profile_data['permissions'] = format(user_permissions, 'd')
    
    return data

validate_user_permissions = flag_validator_factory('user_permissions', Permission)



__all__ = ()

from functools import partial as partial_func

from ..voice_state.fields import (
    put_channel_id, put_deaf, put_mute, validate_channel_id, validate_deaf, validate_mute
)

from .fields import (
    put_avatar_decoration, put_nick, put_role_ids, put_timed_out_until, validate_avatar_decoration,
    validate_nick, validate_role_ids, validate_timed_out_until, validate_timeout_duration
)
from .guild_profile import GUILD_PROFILE_AVATAR, GUILD_PROFILE_BANNER


GUILD_PROFILE_FIELD_CONVERTERS = {
    'deaf': (validate_deaf, put_deaf),
    'mute': (validate_mute, put_mute),
    'nick': (validate_nick, put_nick),
    'role_ids': (validate_role_ids, put_role_ids),
    'roles': (validate_role_ids, put_role_ids),
    'timeout_duration': (validate_timeout_duration, put_timed_out_until),
    'timed_out_until': (validate_timed_out_until, put_timed_out_until),
    'voice_channel': (validate_channel_id, put_channel_id),
    'voice_channel_id': (validate_channel_id, put_channel_id),
}

GUILD_PROFILE_SELF_FIELD_CONVERTERS = {
    'avatar': (
        partial_func(GUILD_PROFILE_AVATAR.validate_icon, allow_data = True),
        partial_func(GUILD_PROFILE_AVATAR.put_into, as_data = True),
    ),
    'avatar_decoration': (validate_avatar_decoration, put_avatar_decoration),
    'banner': (
        partial_func(GUILD_PROFILE_BANNER.validate_icon, allow_data = True),
        partial_func(GUILD_PROFILE_BANNER.put_into, as_data = True),
    ),
    'nick': (validate_nick, put_nick),
}



__all__ = ('AutoModerationActionMetadataTimeout',)

from scarletio import copy_docs

from .base import AutoModerationActionMetadataBase
from .fields import parse_duration, put_duration, validate_duration


class AutoModerationActionMetadataTimeout(AutoModerationActionMetadataBase):
    """
    Timeout action metadata of an auto moderation action.
    
    Attributes
    ----------
    duration : `int`
        The timeout's duration applied on trigger.
    """
    __slots__ = ('duration',)
    
    def __new__(cls, duration = None):
        """
        Creates a new timeout action metadata for ``AutoModerationAction``-s.
        
        Parameters
        ----------
        duration : `None`, `int`, `float` = `None`, Optional
            The timeout's duration applied on trigger.
        
        Raises
        ------
        TypeError
            - If `duration` type is incorrect.
        """
        duration = validate_duration(duration)
        
        self = object.__new__(cls)
        self.duration = duration
        return self
    
    
    @copy_docs(AutoModerationActionMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' duration = ')
        repr_parts.append(repr(self.duration))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @classmethod
    @copy_docs(AutoModerationActionMetadataBase.from_data)
    def from_data(cls, data):
        duration = parse_duration(data)
        
        self = object.__new__(cls)
        self.duration = duration
        return self
    
    
    @copy_docs(AutoModerationActionMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        put_duration(self.duration, data, defaults)
        return data
    
    
    @copy_docs(AutoModerationActionMetadataBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.duration != other.duration:
            return False
        
        return True
    
    
    @copy_docs(AutoModerationActionMetadataBase.__hash__)
    def __hash__(self):
        return self.duration
    
    
    @copy_docs(AutoModerationActionMetadataBase.copy)
    def copy(self):
        new = AutoModerationActionMetadataBase.copy(self)
        
        # duration
        new.duration = self.duration
        
        return new
    
    
    def copy_with(self, *, duration = ...):
        """
        Copies the action metadata and modifies it's attributes by the given values.
        
        Parameters
        ----------
        duration : `None`, `int`, `float`, Optional (Keyword only)
            The timeout's duration applied on trigger.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        if duration is ...:
            duration = self.duration
        else:
            duration = validate_duration(duration)
        
        new = AutoModerationActionMetadataBase.copy(self)
        new.duration = duration
        return new



__all__ = ()

from .constants import PLUGIN_ROOTS


def is_tuple_starting_with(tuple_1, tuple_2):
    """
    Returns whether `tuple_1` starts with `tuple_2`.
    
    Parameters
    ----------
    tuple_1 : `tuple`
        The first tuple which should start with the second one.
    tuple_2 : `tuple`
        The second tuple.
    
    Returns
    -------
    is_tuple_starting_with : `bool`
    """
    tuple_1_length = len(tuple_1)
    tuple_2_length = len(tuple_2)
    
    if tuple_1_length > tuple_2_length:
        return tuple_1[:tuple_2_length] == tuple_2
    
    if tuple_1_length == tuple_2_length:
        return tuple_1 == tuple_2
    
    return False


def register_plugin_root(name):
    """
    Registers a plugin root. Plugin roots might be used to detect nested plugin files.
    
    Parameters
    ----------
    name : `str`
        A plugin's name.
    """
    name_split = tuple(name.split('.'))
    
    for plugin_root in PLUGIN_ROOTS:
        if is_tuple_starting_with(name_split, plugin_root):
            return
    
    to_remove_roots = None
    
    for plugin_root in PLUGIN_ROOTS:
        if is_tuple_starting_with(plugin_root, name_split):
            if to_remove_roots is None:
                to_remove_roots = []
            
            to_remove_roots.append(plugin_root)
    
    if (to_remove_roots is not None):
        for plugin_root in to_remove_roots:
            PLUGIN_ROOTS.discard(plugin_root)
    
    PLUGIN_ROOTS.add(name_split)


def is_in_plugin_root(name):
    """
    Returns whether the given name is inside of a plugin root.
    
    Parameters
    ----------
    name : `str`
        A plugin's name.
    
    Returns
    -------
    is_in_plugin_root : `bool`
    """
    name_split = tuple(name.split('.'))
    for plugin_root in PLUGIN_ROOTS:
        if is_tuple_starting_with(name_split, plugin_root):
            return True
    
    return False



__all__ = ('FormSubmitCommand', )

from scarletio import copy_docs

from .....discord.events.handling_helpers import check_name

from ...converters import get_form_submit_command_parameter_converters
from ...exceptions import handle_command_exception
from ...responding import process_command_coroutine
from ...response_modifier import ResponseModifier
from ...wrappers import CommandWrapper

from ..command_base_custom_id import CommandBaseCustomId
from ..command_base_custom_id.helpers import _validate_custom_ids, _validate_name, split_and_check_satisfaction

from .constants import COMMAND_TARGETS_FORM_COMPONENT_COMMAND


class FormSubmitCommand(CommandBaseCustomId):
    """
    A command, which is called if a message component interaction is received with a matched `custom_id`.
    
    Attributes
    ----------
    _command_function : `async-callable˛
        The command's function to call.
    
    _exception_handlers : `None | list<CoroutineFunction>`
        Exception handlers added with ``.error`` to the interaction handler.
    
    _keyword_parameter_converters : `tuple<ParameterConverter>`
        Parameter converters for keyword parameters.
    
    _multi_parameter_converter : `None | ParameterConverter`
        Parameter converter for positional parameter.
    
    _parent_reference : `None | WeakReferer<SelfReferenceInterface>`
        The parent slasher of the component command.
    
    _parameter_converters : `tuple<ParameterConverter>`
        Parsers to parse command parameters.
    
    _string_custom_ids : `None | tuple<str>`
        The custom id-s to wait for.
    
    _regex_custom_ids : `None | tuple<RegexMatcher>`.
        Regex matchers to match custom-ids.
    
    name : `str`
        The component commands name.
        
        Only used for debugging.

    response_modifier : `None | ResponseModifier`
        Modifies values returned and yielded to command coroutine processor.
    """
    __slots__ = ('_keyword_parameter_converters', '_multi_parameter_converter')
    
    
    def __new__(cls, function, name = None, *, custom_id = ..., target = ..., **keyword_parameters):
        """
        Creates a new form submit command with the given parameters.
        
        Parameters
        ----------
        function : `async-callable`
            The function used as the command when using the respective slash command.
        
        name : `None | str` = `None`, Optional
            The name of the component command.
        
        custom_id : `str | re.Pattern | (list | set)<str | re.Pattern>`, Optional (Keyword only)
            Custom id to match by the component command.
        
        target : `None`, `str`, Optional (Keyword only)
            The component command's target.
        
        Other Parameters
        ----------------
        allowed_mentions : `None | str, UserBase | Role | AllowedMentionProxy | list<str | UserBase | Role> \
                , Optional (Keyword only)
            Which user or role can the response message ping (or everyone).
        
        show_for_invoking_user_only : `bool`, Optional (Keyword only)
            Whether the response message should only be shown for the invoking user.
        
        wait_for_acknowledgement : `bool`, Optional (Keyword only)
            Whether acknowledge tasks should be ensure asynchronously.
        
        Raises
        ------
        TypeError
            If a parameter's type is incorrect.
        ValueError
            If a parameter's value is incorrect.
        """
        if custom_id is ...:
            raise ValueError(
                '`custom_id` parameter is required.'
            )
        
        if (target is not ...) and (target not in COMMAND_TARGETS_FORM_COMPONENT_COMMAND):
            raise ValueError(
                f'`target` can be any of `{COMMAND_TARGETS_FORM_COMPONENT_COMMAND!r}`\'s '
                f'values, got {target!r}.'
            )
        
        if (function is not None) and isinstance(function, CommandWrapper):
            command_function, wrappers = function.fetch_function_and_wrappers_back()
        else:
            command_function = function
            wrappers = None
        
        # Pre validate
        name = _validate_name(name)
        custom_id = _validate_custom_ids(custom_id)
        
        # Check extra parameters
        response_modifier = ResponseModifier(keyword_parameters)
        if keyword_parameters:
            raise TypeError(f'Extra or unused parameters: {keyword_parameters!r}.')
        
        # post validate
        name = check_name(command_function, name)
        command_function, parameter_converters, multi_parameter_converter, keyword_parameter_converters = \
            get_form_submit_command_parameter_converters(command_function)
        string_custom_ids, regex_custom_ids = split_and_check_satisfaction(custom_id, parameter_converters)
        
        # Construct
        self = object.__new__(cls)
        self._command_function = command_function
        self._parameter_converters = parameter_converters
        self._keyword_parameter_converters = keyword_parameter_converters
        self._multi_parameter_converter = multi_parameter_converter
        self._string_custom_ids = string_custom_ids
        self._regex_custom_ids = regex_custom_ids
        self._parent_reference = None
        self._exception_handlers = None
        self.name = name
        self.response_modifier = response_modifier
        
        if (wrappers is not None):
            for wrapper in wrappers:
                wrapper.apply(self)
        
        return self
    
    
    @copy_docs(CommandBaseCustomId.invoke)
    async def invoke(self, client, interaction_event, regex_match):
        # Positional parameters
        positional_parameters = []
        
        for parameter_converter in self._parameter_converters:
            try:
                parameter = await parameter_converter(client, interaction_event, regex_match)
            except GeneratorExit:
                raise
            
            except BaseException as err:
                exception = err
            
            else:
                positional_parameters.append(parameter)
                continue
            
            await handle_command_exception(
                self,
                client,
                interaction_event,
                exception,
            )
            return
        
        parameter_converter = self._multi_parameter_converter
        if (parameter_converter is not None):
            try:
                parameters = await parameter_converter(client, interaction_event, regex_match)
            except GeneratorExit:
                raise
            
            except BaseException as err:
                exception = err
            
            else:
                if (parameters is not None):
                    positional_parameters.extend(parameters)
                
                exception = None
            
            # Call it here to not include the received exception as context
            if (exception is not None):
                await handle_command_exception(
                    self,
                    client,
                    interaction_event,
                    exception,
                )
                return
        
        # Keyword parameters
        keyword_parameters = {}
        
        for parameter_converter in self._keyword_parameter_converters:
            try:
                parameter = await parameter_converter(client, interaction_event, regex_match)
            except GeneratorExit:
                raise
            
            except BaseException as err:
                exception = err
            
            else:
                keyword_parameters[parameter_converter.parameter_name] = parameter
                continue
            
            await handle_command_exception(
                self,
                client,
                interaction_event,
                exception,
            )
            return
        
        # Call command
        command_coroutine = self._command_function(*positional_parameters, **keyword_parameters)
        
        try:
            await process_command_coroutine(client, interaction_event, self.response_modifier, command_coroutine)
        except GeneratorExit:
            raise
        
        except BaseException as err:
            exception = err
        
        else:
            return
        
        await handle_command_exception(
            self,
            client,
            interaction_event,
            exception,
        )
        return
    
    
    @copy_docs(CommandBaseCustomId.copy)
    def copy(self):
        new = CommandBaseCustomId.copy(self)
        
        new._keyword_parameter_converters = self._keyword_parameter_converters
        new._multi_parameter_converter = self._multi_parameter_converter
        
        return new



__all__ = ()


EMBEDDED_ACTIVITY_UPDATE_NONE = 0
EMBEDDED_ACTIVITY_UPDATE_CREATE = 1
EMBEDDED_ACTIVITY_UPDATE_DELETE = 2
EMBEDDED_ACTIVITY_UPDATE_UPDATE = 3
EMBEDDED_ACTIVITY_UPDATE_USER_ADD = 4
EMBEDDED_ACTIVITY_UPDATE_USER_DELETE = 5



__all__ = ()

from .base import ScheduledEventEntityMetadataBase
from .location import ScheduledEventEntityMetadataLocation
from .stage import ScheduledEventEntityMetadataStage


def try_get_scheduled_event_metadata_type_from_data(data):
    """
    Tries to detect what type of scheduled metadata the given data is.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Scheduled event metadata data.
    
    Returns
    -------
    metadata_type : `type<ScheduledEventEntityMetadataBase>`
    """
    if 'location' in data:
        metadata_type = ScheduledEventEntityMetadataLocation
    elif 'speaker_ids' in data:
        metadata_type = ScheduledEventEntityMetadataStage
    else:
        metadata_type = ScheduledEventEntityMetadataBase
    
    return metadata_type



__all__ = ('ApplicationIntegrationTypeConfiguration',)

from scarletio import RichAttributeErrorBaseType

from .fields import parse_install_parameters, put_install_parameters, validate_install_parameters


class ApplicationIntegrationTypeConfiguration(RichAttributeErrorBaseType):
    """
    Integration type specific configuration for installing the application.
    
    Attributes
    ----------
    install_parameters : `None`, ``ApplicationInstallParameters``
        Parameters to install the integration with.        
    """
    __slots__ = ('install_parameters', )
    
    def __new__(cls, *, install_parameters = ...):
        """
        Creates a new application integration type configuration.
        
        Parameters
        ----------
        install_parameters : `None`, ``ApplicationInstallParameters``, Optional (Keyword only)
            Parameters to install the integration with.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        if install_parameters is ...:
            install_parameters = None
        else:
            install_parameters = validate_install_parameters(install_parameters)
        
        self = object.__new__(cls)
        self.install_parameters = install_parameters
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a application integration type configuration instance from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Application installation parameters data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.install_parameters = parse_install_parameters(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the application integration type configuration to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_install_parameters(self.install_parameters, data, defaults)
        return data
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates an integration type configuration with its default values set.
        
        Returns
        -------
        self : `instance<cls>
        """
        self = object.__new__(cls)
        self.install_parameters = None
        return self
    
    
    def __repr__(self):
        """Returns the application install parameters' representation."""
        repr_parts = ['<', type(self).__name__]
        
        install_parameters = self.install_parameters
        if (install_parameters is not None):
            repr_parts.append(', install_parameters = ')
            repr_parts.append(repr(install_parameters))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the application install parameters' hash value."""
        hash_value = 0
        
        # install_parameters
        install_parameters = self.install_parameters
        if (install_parameters is not None):
            hash_value ^= hash(install_parameters)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two application install parameters are equal."""
        if type(other) is not ApplicationIntegrationTypeConfiguration:
            return NotImplemented
        
        if self.install_parameters != other.install_parameters:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the application install parameters.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        install_parameters = self.install_parameters
        if (install_parameters is not None):
            install_parameters = install_parameters.copy()
        new.install_parameters = install_parameters
        
        return new

    
    def copy_with(self, install_parameters = ...):
        """
        Copies the application install parameters with the given fields.
        
        Parameters
        ----------
        install_parameters : `None`, ``ApplicationInstallParameters``, Optional (Keyword only)
            Parameters to install the integration with.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        if install_parameters is ...:
            install_parameters = self.install_parameters
        else:
            install_parameters = validate_install_parameters(install_parameters)
        
        new = object.__new__(type(self))
        new.install_parameters = install_parameters
        return new



__all__ = ('Entitlement',)

from scarletio import export

from ...bases import DiscordEntity
from ...core import ENTITLEMENTS
from ...precreate_helpers import process_precreate_parameters_and_raise_extra

from .fields import (
    parse_application_id, parse_consumed, parse_deleted, parse_ends_at, parse_guild_id, parse_id, parse_sku_id,
    parse_starts_at, parse_subscription_id, parse_type, parse_user_id, put_application_id, put_consumed,
    put_deleted, put_ends_at, put_guild_id, put_id, put_owner_id, put_owner_type,
    put_sku_id, put_starts_at, put_subscription_id, put_type, put_user_id,
    validate_application_id, validate_consumed, validate_deleted, validate_ends_at, validate_guild_id, validate_id,
    validate_sku_id, validate_starts_at, validate_subscription_id, validate_type, validate_user_id
)
from .preinstanced import EntitlementOwnerType, EntitlementType


PRECREATE_FIELDS = {
    'application': ('application_id', validate_application_id),
    'application_id': ('application_id', validate_application_id),
    'consumed': ('consumed', validate_consumed),
    'deleted': ('deleted', validate_deleted),
    'ends_at': ('ends_at', validate_ends_at),
    'entitlement_type': ('type', validate_type),
    'guild': ('guild_id', validate_guild_id),
    'guild_id': ('guild_id', validate_guild_id),
    'sku': ('sku_id', validate_sku_id),
    'sku_id': ('sku_id', validate_sku_id),
    'starts_at': ('starts_at', validate_starts_at),
    'subscription': ('subscription_id', validate_subscription_id),
    'subscription_id': ('subscription_id', validate_subscription_id),
    'user': ('user_id', validate_user_id),
    'user_id': ('user_id', validate_user_id),
}


@export
class Entitlement(DiscordEntity):
    """
    Represent that a user or guild has access to a premium offer.
    
    Attributes
    ----------
    application_id : `int`
        The entitlement's owner application's identifier.
    
    consumed : `bool`
        Whether the entitlement is already consumed.
        Not applicable for subscriptions.
    
    deleted : `bool`
        Whether the entitlement is deleted.
    
    ends_at : `None | DateTime`
        When the entitlement ends.
    
    guild_id : `int`
        The guild's identifier that was granted access to the stock keeping unit.
    
    id : `int`
        The unique identifier number of the entitlement.
    
    sku_id : `int`
        The stock keeping unit's identifier the this entitlement grants access to.
    
    starts_at : `None | DateTime`
        When the entitlement starts.
    
    subscription_id : `int`
        The subscription's identifier the entitlement is part of.
    
    type : ``EntitlementType``
        The entitlement's type.
    
    user_id : `int`
        The user's identifier that was granted access to the stock keeping unit.
    
    Notes
    -----
    Entitlement instances are weakreferable.
    """
    __slots__ = (
        '__weakref__', 'application_id', 'consumed', 'deleted', 'ends_at', 'guild_id', 'sku_id', 'starts_at',
        'subscription_id', 'type', 'user_id'
    )
    
    def __new__(cls, *, guild_id = ..., sku_id = ..., user_id = ...):
        """
        Creates a new partial entitlement.
        
        Parameters
        ----------
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The guild's identifier that was granted access to the stock keeping unit.
    
        sku_id : `int`, ``SKU``, Optional (Keyword only)
            The stock keeping unit's identifier the this entitlement grants access to.
        
        user_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The user's identifier that was granted access to the stock keeping unit.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # sku_id
        if sku_id is ...:
            sku_id = 0
        else:
            sku_id = validate_sku_id(sku_id)
        
        # user_id
        if user_id is ...:
            user_id = 0
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        self = object.__new__(cls)
        self.application_id = 0
        self.consumed = False
        self.deleted = False
        self.ends_at = None
        self.guild_id = guild_id
        self.id = 0
        self.sku_id = sku_id
        self.starts_at = None
        self.subscription_id = 0
        self.type = EntitlementType.none
        self.user_id = user_id
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new entitlement from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entitlement data.
        
        Returns
        -------
        new : `instance<cls>`
        """
        entitlement_id = parse_id(data)
        
        try:
            self = ENTITLEMENTS[entitlement_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = entitlement_id
            self._set_attributes(data)
            ENTITLEMENTS[entitlement_id] = self
        else:
            self._set_attributes(data)
        
        return self
    
    
    @classmethod
    def from_data_is_created(cls, data):
        """
        Creates a new entitlement 
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entitlement data.
        
        Returns
        -------
        new : `instance<cls>`
        created : `bool`
        """
        entitlement_id = parse_id(data)
        
        try:
            self = ENTITLEMENTS[entitlement_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = entitlement_id
            self._set_attributes(data)
            ENTITLEMENTS[entitlement_id] = self
            created = True
        else:
            self._set_attributes(data)
            created = False
        
        return self, created
    
    
    def _set_attributes(self, data):
        """
        Sets the entitlement's attributes from the given data. (Except `.id`.)
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entitlement data.
        """
        self.application_id = parse_application_id(data)
        self.guild_id = parse_guild_id(data)
        self.sku_id = parse_sku_id(data)
        self.subscription_id = parse_subscription_id(data)
        self.type = parse_type(data)
        self.user_id = parse_user_id(data)
        
        self._update_attributes(data)
    
    
    def _update_attributes(self, data):
        """
        Updates the entitlement by overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entitlement data.
        """
        self.consumed = parse_consumed(data)
        self.deleted = parse_deleted(data)
        self.ends_at = parse_ends_at(data)
        self.starts_at = parse_starts_at(data)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the attributes of the entitlement and returns the changed ones within an `attribute-name` -
        `old-value` relation.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Entitlement data.
        
        Returns
        -------
        old_attributes : `dict<str, object>`
            The updated attributes.
            
            The returned dictionary might contain the following items:
            
            +---------------------------+-----------------------------------------------+
            | Key                       | Value                                         |
            +===========================+===============================================+
            | consumed                  | `bool`                                        |
            +---------------------------+-----------------------------------------------+
            | deleted                   | `bool`                                        |
            +---------------------------+-----------------------------------------------+
            | ends_at                   | `None`, `DateTime`                            |
            +---------------------------+-----------------------------------------------+
            | starts_at                 | `None`, `DateTime`                            |
            +---------------------------+-----------------------------------------------+
        """
        old_attributes = {}
        
        # consumed
        consumed = parse_consumed(data)
        if self.consumed != consumed:
            old_attributes['consumed'] = self.consumed
            self.consumed = consumed
        
        # deleted
        deleted = parse_deleted(data)
        if self.deleted != deleted:
            old_attributes['deleted'] = self.deleted
            self.deleted = deleted
        
        # ends_at
        ends_at = parse_ends_at(data)
        if self.ends_at != ends_at:
            old_attributes['ends_at'] = self.ends_at
            self.ends_at = ends_at
        
        # starts_at
        starts_at = parse_starts_at(data)
        if self.starts_at != starts_at:
            old_attributes['starts_at'] = self.starts_at
            self.starts_at = starts_at
        
        return old_attributes
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the entitlement into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_sku_id(self.sku_id, data, defaults)
        
        if include_internals:
            put_application_id(self.application_id, data, defaults)
            put_consumed(self.consumed, data, defaults)
            put_deleted(self.deleted, data, defaults)
            put_ends_at(self.ends_at, data, defaults)
            put_guild_id(self.guild_id, data, defaults)
            put_id(self.id, data, defaults)
            put_starts_at(self.starts_at, data, defaults)
            put_subscription_id(self.subscription_id, data, defaults)
            put_type(self.type, data, defaults)
            put_user_id(self.user_id, data, defaults)
        else:
            put_owner_id(self.owner_id, data, defaults)
            put_owner_type(self.owner_type, data, defaults)
        
        return data
    
    
    @classmethod
    def _create_empty(cls, entitlement_id):
        """
        Creates a new entitlement instance with it's attribute set to their default values.
        
        Parameters
        ----------
        entitlement_id : `int`
            The entitlement's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.application_id = 0
        self.consumed = False
        self.deleted = False
        self.ends_at = None
        self.guild_id = 0
        self.id = entitlement_id
        self.sku_id = 0
        self.starts_at = None
        self.subscription_id = 0
        self.type = EntitlementType.none
        self.user_id = 0
        return self
    

    @classmethod
    def precreate(cls, entitlement_id, **keyword_parameters):
        """
        Creates an entitlement instance.
        
        Parameters
        ----------
        entitlement_id : `int`
            The entitlement's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        application : `int`, ``Application``, Optional (Keyword only)
            Alternative for `application_id`.
        
        application_id : `int`, ``Application``, Optional (Keyword only)
            The entitlement's owner application's identifier.
        
        consumed : `bool`, Optional (Keyword only)
            Whether the entitlement is already consumed.
        
        deleted : `bool`, Optional (Keyword only)
            Whether the entitlement is deleted.
        
        ends_at : `None`, `DateTime`, Optional (Keyword only)
            When the entitlement ends.
        
        entitlement_type : ``EntitlementType``, `int`, Optional (Keyword only)
            The entitlement's type.
        
        guild : `int`, ``Guild``, Optional (Keyword only)
            Alternative for `guild_id`.
        
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The guild's identifier that was granted access to the stock keeping unit.
        
        sku : `int`, ``SKU``, Optional (Keyword only)
            Alternative for `sku_id`.
        
        sku_id : `int`, ``SKU``, Optional (Keyword only)
            The stock keeping unit's identifier the this entitlement grants access to.
    
        starts_at : `None`, `DateTime`, Optional (Keyword only)
            When the entitlement starts.
        
        subscription : `int`, Optional (Keyword only)
            Alternative for `subscription_id`.
        
        subscription_id : `int`, Optional (Keyword only)
            The subscription's identifier the entitlement is part of.
        
        user : `int`, ``ClientUserBase``, Optional (Keyword only)
            Alternative for `user_id`.
        
        user_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The user's identifier that was granted access to the stock keeping unit.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        entitlement_id = validate_id(entitlement_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = ENTITLEMENTS[entitlement_id]
        except KeyError:
            self = cls._create_empty(entitlement_id)
            ENTITLEMENTS[entitlement_id] = self
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    def __repr__(self):
        """Returns the entitlement's representation."""
        repr_parts = ['<', type(self).__name__]
        
        entitlement_id = self.id
        if entitlement_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(self.id))
            repr_parts.append(',')
        
        repr_parts.append(' sku_id = ')
        repr_parts.append(repr(self.sku_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two entitlements are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two entitlements are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self is equal to other. Other must be same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self.id != other.id:
                return False
        
        # application_id -> ignore, internal
        # consumed -> ignore, internal
        # deleted -> ignore, internal
        # ends_at -> ignore, internal
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        # sku_id
        if self.sku_id != other.sku_id:
            return False
        
        # starts_at -> ignore, internal
        # subscription_id -> ignore, internal
        # type -> ignore, internal
        
        # user_id
        if self.user_id != other.user_id:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the entitlement."""
        entitlement_id = self.id
        if entitlement_id:
            return entitlement_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Calculates the entitlement's hash based on their fields.
        
        This method is called by ``.__hash__`` if the channel has no ``.id`` set.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # application_id -> ignore, internal
        # consumed -> ignore, internal
        # deleted -> ignore, internal
        # ends_at -> ignore, internal
        
        # guild_id
        hash_value ^= self.guild_id
        
        # id -> ignore, internal
        
        # sku_id
        hash_value ^= self.sku_id
        
        # starts_at -> ignore, internal
        # subscription_id -> ignore, internal
        # type -> ignore, internal
        
        # user_id
        hash_value ^= self.user_id
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the entitlement.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.application_id = 0
        new.consumed = False
        new.deleted = False
        new.ends_at = None
        new.guild_id = self.guild_id
        new.id = 0
        new.sku_id = self.sku_id
        new.starts_at = None
        new.subscription_id = 0
        new.type = EntitlementType.none
        new.user_id = self.user_id
        return new
    
    
    def copy_with(self, *, guild_id = ..., sku_id = ..., user_id = ...):
        """
        Copies the entitlement with the given fields.
        
        Parameters
        ----------
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The guild's identifier that was granted access to the stock keeping unit.
    
        sku_id : `int`, ``SKU``, Optional (Keyword only)
            The stock keeping unit's identifier the this entitlement grants access to.
        
        user_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The user's identifier that was granted access to the stock keeping unit.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # sku_id
        if sku_id is ...:
            sku_id = self.sku_id
        else:
            sku_id = validate_sku_id(sku_id)
        
        # user_id
        if user_id is ...:
            user_id = self.user_id
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        new = object.__new__(type(self))
        new.application_id = 0
        new.consumed = False
        new.deleted = False
        new.ends_at = None
        new.guild_id = guild_id
        new.id = 0
        new.sku_id = sku_id
        new.starts_at = None
        new.subscription_id = 0
        new.type = EntitlementType.none
        new.user_id = user_id
        return new
    
    
    @property
    def partial(self):
        """
        Returns whether the entity is partial.
        
        Returns
        -------
        partial : `bool
        """
        return (self.id == 0)
    
    
    @property
    def owner_id(self):
        """
        Returns the entitlement's owner's identifier.
        It is either its ``.guild_id`` or ``.user_id``.
        
        Returns
        -------
        owner_id : `int`
        """
        guild_id = self.guild_id
        if guild_id:
            return guild_id
        
        user_id = self.user_id
        if user_id:
            return user_id
        
        return 0
    
    
    @property
    def owner_type(self):
        """
        Returns the entitlement's owner type.
        
        Returns
        -------
        owner_type : ``EntitlementOwnerType``
        """
        if  self.guild_id:
            return EntitlementOwnerType.guild
        
        if self.user_id:
            return EntitlementOwnerType.user
        
        return EntitlementOwnerType.none



__all__ = ('CONVERSION_REPLY_MESSAGE_ID',)

from ....builder.constants import CONVERSION_KIND_FIELD
from ....builder.conversion import Conversion

from ..message_reference_configuration import MessageReferenceConfiguration

from .message_reference_configuration import CONVERSION_MESSAGE_REFERENCE_CONFIGURATION


class CONVERSION_REPLY_MESSAGE_ID(Conversion):
    # Generic
    
    name = 'reply_message_id'
    name_aliases = None
    expected_types_messages = '`int`'
    kind = CONVERSION_KIND_FIELD
    output_conversion = CONVERSION_MESSAGE_REFERENCE_CONFIGURATION
    
    # Setting
    
    set_merger = None
    set_type = None
    set_type_processor = None
    set_identifier = None
    set_listing_identifier = None
    
    
    def set_validator(value):
        # None
        if value is None:
            yield MessageReferenceConfiguration()
            return
        
        # int
        if isinstance(value, int):
            yield MessageReferenceConfiguration(message_id = value)
            return
        
        # No other cases
        return
    
    # Reading
    
    get_default = 0
    
    def get_processor(value):
        if value is None:
            return 0
        
        return value.message_id
    
    # Serialization
    
    serializer_key = None
    serializer_optional = None
    serializer_required = None
    
    # Sorting
    
    sort_priority = 2400



__all__ = ('SKU',)

from scarletio import export

from ...bases import DiscordEntity
from ...core import SKUS
from ...precreate_helpers import process_precreate_parameters_and_raise_extra

from .fields import (
    parse_access_type, parse_application_id, parse_features, parse_flags, parse_id, parse_name, parse_premium,
    parse_release_at, parse_slug, parse_type, put_access_type, put_application_id, put_features,
    put_flags, put_id, put_name, put_premium, put_release_at, put_slug, put_type,
    validate_access_type, validate_application_id, validate_features, validate_flags, validate_id, validate_name,
    validate_premium, validate_release_at, validate_slug, validate_type
)
from .flags import SKUFlag
from .preinstanced import SKUAccessType, SKUType


PRECREATE_FIELDS = {
    'access_type': ('access_type', validate_access_type),
    'application': ('application_id', validate_application_id),
    'application_id': ('application_id', validate_application_id),
    'features': ('features', validate_features),
    'flags': ('flags', validate_flags),
    'name': ('name', validate_name),
    'premium': ('premium', validate_premium),
    'release_at': ('release_at', validate_release_at),
    'slug': ('slug', validate_slug),
    'sku_type': ('type', validate_type),
}


@export
class SKU(DiscordEntity):
    """
    A stock keeping unit.
    
    Attributes
    ----------
    access_type : ``SKUAccessType``
        What kind of access the stock keeping unit provides for its content.
    
    application_id : `int`
        The stock keeping unit's owner application identifier.
    
    features : `None, tuple<SKUFeature>`
        The features of the stock keeping unit.
    
    flags : ``SKUFlag``
        The flags of the stock keeping unit.
    
    id : `int`
        The unique identifier number of the stock keeping unit.
    
    name : `str`
        The name of the stock keeping unit.
    
    premium : `bool`
        Whether the stock keeping unit is a premium one.
    
    release_at : `None`, `DateTime`
        When the stock keeping unit has its release.
        Can be both in the past and in the future as well.
    
    slug : `None`, `str`
        System generated url to the stock keeping unit generated based on its name.
    
    type : ``SKUType``
        The stock keeping unit's type.
    
    Notes
    -----
    Stock keeping unit instances are weakreferable.
    """
    __slots__ = (
        '__weakref__', 'access_type', 'application_id', 'features', 'flags', 'name', 'premium', 'release_at', 'slug',
        'type'
    )
    
    def __new__(
        cls,
        *,
        access_type = ...,
        features = ...,
        flags = ...,
        name = ...,
        premium = ...,
        release_at = ...,
        sku_type = ...,
    ):
        """
        Creates a new partial stock keeping unit.
        
        Parameters
        ----------
        access_type : ``SKUAccessType``, `int`, Optional (Keyword only)
            What kind of access the stock keeping unit provides for its content.
        
        features : `None`, `iterable<SKUFeature | int>`, `SKUFeature`, `int`, Optional (Keyword only)
            The features of the stock keeping unit.
        
        flags : ``SKUFlag``, `int`, Optional (Keyword only)
            The flags of the stock keeping unit.
        
        name : `str`, Optional (Keyword only)
            The name of the stock keeping unit.
        
        premium : `bool`, Optional (Keyword only)
            Whether the stock keeping unit is a premium one.
            
        release_at : `None`, `DateTime`, Optional (Keyword only)
            When the stock keeping unit has its release.
        
        sku_type : ``SKUType``, `int`, Optional (Keyword only)
            The stock keeping unit's type.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # access_type
        if access_type is ...:
            access_type = SKUAccessType.none
        else:
            access_type = validate_access_type(access_type)
        
        # features
        if features is ...:
            features = None
        else:
            features = validate_features(features)
        
        # flags
        if flags is ...:
            flags = SKUFlag()
        else:
            flags = validate_flags(flags)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # premium
        if premium is ...:
            premium = False
        else:
            premium = validate_premium(premium)
        
        # release_at
        if release_at is ...:
            release_at = None
        else:
            release_at = validate_release_at(release_at)
        
        # sku_type
        if sku_type is ...:
            sku_type = SKUType.none
        else:
            sku_type = validate_type(sku_type)
        
        # Construct
        self = object.__new__(cls)
        self.access_type = access_type
        self.application_id = 0
        self.features = features
        self.flags = flags
        self.id = 0
        self.name = name
        self.premium = premium
        self.release_at = release_at
        self.slug = None
        self.type = sku_type
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new stock keeping unit.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Stock keeping unit data.
        
        Returns
        -------
        new : `instance<cls>`
        """
        sku_id = parse_id(data)
        try:
            self = SKUS[sku_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = sku_id
            self._set_attributes(data)
            SKUS[sku_id] = self
        else:
            self._set_attributes(data)
        
        return self
    
    def _set_attributes(self, data):
        """
        Sets the stock keeping unit's attributes. (Except `.id`.)
        
        Parameters
        ----------
        data : `dict<str, object>`
            Stock keeping unit data.
        """
        self.access_type = parse_access_type(data)
        self.application_id = parse_application_id(data)
        self.features = parse_features(data)
        self.flags = parse_flags(data)
        self.name = parse_name(data)
        self.premium = parse_premium(data)
        self.release_at = parse_release_at(data)
        self.slug = parse_slug(data)
        self.type = parse_type(data)
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the stock keeping unit into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        
        put_access_type(self.access_type, data, defaults)
        put_features(self.features, data, defaults)
        put_flags(self.flags, data, defaults)
        put_name(self.name, data, defaults)
        put_premium(self.premium, data, defaults)
        put_release_at(self.release_at, data, defaults)
        put_type(self.type, data, defaults)
        
        if include_internals:
            put_application_id(self.application_id, data, defaults)
            put_id(self.id, data, defaults)
            put_slug(self.slug, data, defaults)
        
        return data
    
    
    @classmethod
    def _create_empty(cls, sku_id):
        """
        Creates a new stock keeping unit instance with it's attribute set to their default values.
        
        Parameters
        ----------
        sku_id : `int`
            The stock keeping unit's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.access_type = SKUAccessType.none
        self.application_id = 0
        self.features = None
        self.flags = SKUFlag()
        self.id = sku_id
        self.name = ''
        self.premium = False
        self.release_at = None
        self.slug = None
        self.type = SKUType.none
        return self
    

    @classmethod
    def precreate(cls, sku_id, **keyword_parameters):
        """
        Creates an stock keeping unit instance.
        
        Parameters
        ----------
        sku_id : `int`
            The stock keeping unit's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        access_type : ``SKUAccessType``, `int`, Optional (Keyword only)
            What kind of access the stock keeping unit provides for its content.
        
        application : `int`, ``Application``, Optional (Keyword only)
            Alternative for `application_id`.
        
        application_id : `int`, ``Application``, Optional (Keyword only)
            The stock keeping unit's owner application identifier.
        
        features : `None`, `iterable<SKUFeature | int>`, `SKUFeature`, `int`, Optional (Keyword only)
            The features of the stock keeping unit.
        
        flags : ``SKUFlag``, `int`, Optional (Keyword only)
            The flags of the stock keeping unit.
        
        name : `str`, Optional (Keyword only)
            The stock keeping unit's name.
        
        premium : `bool`, Optional (Keyword only)
            Whether the stock keeping unit is a premium one.
            
        release_at : `None`, `DateTime`, Optional (Keyword only)
            When the stock keeping unit has its release.
        
        slug : `None`, `str`, Optional (Keyword only)
             System generated url to the stock keeping unit generated based on its name.
        
        sku_type : ``SKUType``, `int`, Optional (Keyword only)
            The stock keeping unit's type.
         
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        sku_id = validate_id(sku_id)

        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = SKUS[sku_id]
        except KeyError:
            self = cls._create_empty(sku_id)
            SKUS[sku_id] = self
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    def __repr__(self):
        """Returns the stock keeping unit's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        sku_id = self.id
        if sku_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(self.id))
            repr_parts.append(',')
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two stock keeping units are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two stock keeping units are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self is equal to other. Other must be same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self.id != other.id:
                return False
        
        # access_type
        if self.access_type is not other.access_type:
            return False
        
        # application_id -> ignore, internal
        
        # features
        if self.features != other.features:
            return False
        
        # flags
        if self.flags != other.flags:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # premium
        if self.premium != other.premium:
            return False
        
        # release_at
        if self.release_at != other.release_at:
            return False
        
        # slug -> ignore, internal
        
        # type
        if self.type is not other.type:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the stock keeping unit."""
        sku_id = self.id
        if sku_id:
            return sku_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Calculates the stock keeping unit's hash based on their fields.
        
        This method is called by ``.__hash__`` if the channel has no ``.id`` set.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # access_type
        hash_value ^= self.access_type.value
        
        # application_id -> ignore, internal
        
        # features
        features = self.features
        if (features is not None):
            hash_value ^= hash(features)
        
        # flags
        hash_value ^= self.flags << 4
        
        # id -> ignore internal
        
        # name
        hash_value ^= hash(self.name)
        
        # premium
        hash_value ^= self.premium << 8
        
        # release_at
        release_at = self.release_at
        if (release_at is not None):
            hash_value ^= hash(release_at)
        
        # slug -> ignore, internal
        
        # type
        hash_value ^= self.type.value << 12
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the stock keeping unit.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.access_type = self.access_type
        new.application_id = 0
        features = self.features
        if (features is not None):
            features = (*features,)
        new.features = features
        new.flags = self.flags
        new.id = 0
        new.name = self.name
        new.premium = self.premium
        new.release_at = self.release_at
        new.slug = None
        new.type = self.type
        return new
    
    
    def copy_with(
        self,
        *,
        access_type = ...,
        features = ...,
        flags = ...,
        name = ...,
        premium = ...,
        release_at = ...,
        sku_type = ...,
    ):
        """
        Copies the stock keeping unit with the given fields.
        
        Parameters
        ----------
        access_type : ``SKUAccessType``, `int`, Optional (Keyword only)
            What kind of access the stock keeping unit provides for its content.
        
        features : `None`, `iterable<SKUFeature | int>`, `SKUFeature`, `int`, Optional (Keyword only)
            The features of the stock keeping unit.
        
        flags : ``SKUFlag``, `int`, Optional (Keyword only)
            The flags of the stock keeping unit.
        
        name : `str`, Optional (Keyword only)
            The stock keeping unit's name.
        
        premium : `bool`, Optional (Keyword only)
            Whether the stock keeping unit is a premium one.
            
        release_at : `None`, `DateTime`, Optional (Keyword only)
            When the stock keeping unit has its release.
        
        sku_type : ``SKUType``, `int`, Optional (Keyword only)
            The stock keeping unit's type.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # access_type
        if access_type is ...:
            access_type = self.access_type
        else:
            access_type = validate_access_type(access_type)
        
        # features
        if features is ...:
            features = self.features
            if (features is not None):
                features = (*features,)
        else:
            features = validate_features(features)
        
        # flags
        if flags is ...:
            flags = self.flags
        else:
            flags = validate_flags(flags)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # premium
        if premium is ...:
            premium = self.premium
        else:
            premium = validate_premium(premium)
        
        # release_at
        if release_at is ...:
            release_at = self.release_at
        else:
            release_at = validate_release_at(release_at)
        
        # sku_type
        if sku_type is ...:
            sku_type = self.type
        else:
            sku_type = validate_type(sku_type)
        
        # Construct
        new = object.__new__(type(self))
        new.access_type = access_type
        new.application_id = 0
        new.features = features
        new.flags = flags
        new.id = 0
        new.name = name
        new.premium = premium
        new.release_at = release_at
        new.slug = None
        new.type = sku_type
        return new
    
    
    @property
    def partial(self):
        """
        Returns whether the stock keeping unit is partial.
        
        Returns
        -------
        partial : `bool
        """
        return (self.id == 0)
    
    
    def has_feature(self, feature):
        """
        Returns whether the stock keeping unit has the give feature.
        
        Parameters
        ----------
        feature : ``SKUFeature``
            The feature to look for.
        
        Returns
        -------
        has_feature : `bool`
        """
        features = self.features
        if features is None:
            return False
        
        return feature in features
    
    
    def iter_features(self):
        """
        Iterates over the features of the stock keeping unit.
        
        This method is an iterable generator.
        
        Yields
        ------
        feature : ``SKUFeature``
        """
        features = self.features
        if (features is not None):
            yield from features



__all__ = ()

from ...field_parsers import (
    int_parser_factory, nullable_array_parser_factory, nullable_date_time_parser_factory,
    nullable_entity_array_parser_factory, preinstanced_array_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    int_optional_putter_factory, int_putter_factory, nullable_date_time_optional_putter_factory,
    nullable_entity_array_optional_putter_factory, nullable_value_array_optional_putter_factory,
    preinstanced_array_optional_putter_factory, preinstanced_putter_factory
)
from ...field_validators import (
    int_conditional_validator_factory, nullable_date_time_validator_factory, nullable_entity_array_validator_factory,
    nullable_sorted_int_array_conditional_validator_factory, preinstanced_array_validator_factory,
    preinstanced_validator_factory
)

from ..schedule_nth_weeks_day import ScheduleNthWeeksDay, ScheduleWeeksDay

from .preinstanced import ScheduleFrequency, ScheduleMonth


# by_month_days

parse_by_month_days = nullable_array_parser_factory('by_month_day')
put_by_month_days = nullable_value_array_optional_putter_factory('by_month_day')
validate_by_month_days = nullable_sorted_int_array_conditional_validator_factory(
    'by_month_days',
    (lambda month_day : month_day >= 1 and month_day <= 31),
    '>= 1 and <= 31',
)


# by_months

parse_by_months = preinstanced_array_parser_factory('by_month', ScheduleMonth)
put_by_months = preinstanced_array_optional_putter_factory('by_month')
validate_by_months = preinstanced_array_validator_factory('by_months', ScheduleMonth)


# by_nth_weeks_days

parse_by_nth_weeks_days = nullable_entity_array_parser_factory('by_n_weekday', ScheduleNthWeeksDay)
put_by_nth_weeks_days = nullable_entity_array_optional_putter_factory('by_n_weekday', ScheduleNthWeeksDay)
validate_by_nth_weeks_days = nullable_entity_array_validator_factory('by_nth_weeks_days', ScheduleNthWeeksDay)


# by_weeks_days

parse_by_weeks_days = preinstanced_array_parser_factory('by_weekday', ScheduleWeeksDay)
put_by_weeks_days = preinstanced_array_optional_putter_factory('by_weekday')
validate_by_weeks_days = preinstanced_array_validator_factory('by_weeks_days', ScheduleWeeksDay)


# by_year_days

parse_by_year_days = nullable_array_parser_factory('by_year_day')
put_by_year_days = nullable_value_array_optional_putter_factory('by_year_day')
validate_by_year_days = nullable_sorted_int_array_conditional_validator_factory(
    'by_year_days',
    (lambda year_day : year_day >= 1 and year_day <= 366),
    '>= 1 and <= 366',
)


# end

parse_end = nullable_date_time_parser_factory('end')
put_end = nullable_date_time_optional_putter_factory('end')
validate_end = nullable_date_time_validator_factory('end')


# frequency

parse_frequency = preinstanced_parser_factory('frequency', ScheduleFrequency, ScheduleFrequency.yearly)
put_frequency = preinstanced_putter_factory('frequency')
validate_frequency = preinstanced_validator_factory('frequency', ScheduleFrequency)


# occurrence_count_limit

parse_occurrence_count_limit = int_parser_factory('count', 0)
put_occurrence_count_limit = int_optional_putter_factory('count', 0)
validate_occurrence_count_limit = int_conditional_validator_factory(
    'occurrence_count_limit',
    0,
    (lambda count : count >= 0),
    f'>= 0'
)


# occurrence_spacing

parse_occurrence_spacing = int_parser_factory('interval', 1)
put_occurrence_spacing = int_putter_factory('interval')
validate_occurrence_spacing = int_conditional_validator_factory(
    'occurrence_spacing',
    1,
    (lambda interval : interval >= 1),
    f'>= 1'
)


# start

parse_start = nullable_date_time_parser_factory('start')
put_start = nullable_date_time_optional_putter_factory('start')
validate_start = nullable_date_time_validator_factory('start')



__all__ = ('ActivityType',)

from warnings import warn

from scarletio import class_property

from ...bases import Preinstance as P, PreinstancedBase

from ..activity_metadata import (
    ActivityMetadataBase, ActivityMetadataCustom, ActivityMetadataHanging, ActivityMetadataRich
)


class ActivityType(PreinstancedBase, value_type = int):
    """
    Represents an ``Activity``'s type.
    
    Attributes
    ----------
    value : `int`
        The Discord side identifier value of the activity type.
    
    name : `str`
        The default name of the activity type.
    
    metadata_type : `type<ActivityMetadataBase>`
        The activity type's respective metadata type.
    
    Type Attributes
    ----------------
    Every predefined activity type is also stored as a type attribute:
    
    +-----------------------+-----------------------+-----------+---------------------------------------+
    | Type attribute name   | Name                  | Value     | Metadata type                         |
    +=======================+=======================+===========+=======================================+
    | unknown               | unknown               | -1        | ``ActivityMetadataBase``              |
    +-----------------------+-----------------------+-----------+---------------------------------------+
    | playing               | playing               | 0         | ``ActivityMetadataRich``              |
    +-----------------------+-----------------------+-----------+---------------------------------------+
    | stream                | stream                | 1         | ``ActivityMetadataRich``              |
    +-----------------------+-----------------------+-----------+---------------------------------------+
    | spotify               | spotify               | 2         | ``ActivityMetadataRich``              |
    +-----------------------+-----------------------+-----------+---------------------------------------+
    | watching              | watching              | 3         | ``ActivityMetadataRich``              |
    +-----------------------+-----------------------+-----------+---------------------------------------+
    | custom                | custom                | 4         | ``ActivityMetadataCustom``            |
    +-----------------------+-----------------------+-----------+---------------------------------------+
    | competing             | competing             | 5         | ``ActivityMetadataRich``              |
    +-----------------------+-----------------------+-----------+---------------------------------------+
    | hanging               | hanging               | 6         | ``ActivityMetadataHanging``           |
    +-----------------------+-----------------------+-----------+---------------------------------------+
    """
    __slots__ = ('metadata_type',)
    
    def __new__(cls, value, name = None, metadata_type = None):
        """
        Creates an new activity type.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the activity type.
        
        name : `None | str` = `None`, Optional
            The name of the activity type.
        
        metadata_type : `None | type<ActivityMetadataBase>` = `None`, Optional
            The activity type's respective metadata type.
        """
        if metadata_type is None:
            metadata_type = ActivityMetadataRich
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.metadata_type = metadata_type
        return self
    
    
    # predefined
    unknown = P(-1, 'unknown', ActivityMetadataBase)
    playing = P(0, 'playing', ActivityMetadataRich)
    stream = P(1, 'stream message', ActivityMetadataRich)
    spotify = P(2, 'spotify', ActivityMetadataRich)
    watching = P(3, 'watching', ActivityMetadataRich)
    custom = P(4, 'custom', ActivityMetadataCustom)
    competing = P(5, 'competing', ActivityMetadataRich)
    hanging = P(6, 'hanging', ActivityMetadataHanging)
    
    
    @class_property
    def game(cls):
        """
        Deprecated and will be removed n 2025 September.
        Please use `.playing` instead.
        """
        warn(
            (
                f'`{cls.__name__}.game` is deprecated and will be removed in 2025 September. '
                f'Please use `.playing` instead.'
            ),
            FutureWarning,
            stacklevel = 3,
        )
        
        return cls.playing



__all__ = ()

from functools import partial as partial_func

from scarletio import CallableAnalyzer, RichAttributeErrorBaseType


def validate_exception_handler(function):
    """
    Tests whether the given exception handler accepts the expected amount of parameters.
    
    Parameters
    ----------
    function : `CoroutineFunctionType`
        A function, which handles an exception and returns whether handled it.
        
        The following parameters are passed to it:
        
        +-------------------+-------------------------------------------------------------------------------+
        | Name              | Type                                                                          |
        +===================+===============================================================================+
        | client            | ``Client``                                                                    |
        +-------------------+-------------------------------------------------------------------------------+
        | interaction_event | ``InteractionEvent``                                                          |
        +-------------------+-------------------------------------------------------------------------------+
        | command           | ``ComponentCommand``, ``EmbeddedActivityLaunchCommand``, ``SlashCommand``,    |
        |                   | ``SlashCommandFunction``, ``SlashCommandCategory``,                           |
        |                   | ``SlashCommandParameterAutoCompleter``                                        |
        +-------------------+-------------------------------------------------------------------------------+
        | exception         | `BaseException`                                                               |
        +-------------------+-------------------------------------------------------------------------------+
        
        Should return the following parameters:
        
        +-------------------+-----------+
        | Name              | Type      |
        +===================+===========+
        | handled           | `bool`    |
        +-------------------+-----------+
    
    Returns
    -------
    exception_handler : `CoroutineFunctionType`
    
    Raises
    ------
    TypeError
        - If `exception_handler` accepts bad amount of parameters.
        - If `exception_handler` is not a coroutine function.
    """
    analyzer = CallableAnalyzer(function)
    if not analyzer.is_async():
        raise TypeError(
            f'`exception_handler` can be `async` function, got {function!r}.'
        )
    
    min_, max_ = analyzer.get_non_reserved_positional_parameter_range()
    if min_ > 4:
        raise TypeError(
            f'`exception_handler` should accept `4` parameters, meanwhile the given callable expects at '
            f'least `{min_!r}`, got {function!r}.'
        )
    
    if min_ != 4:
        if max_ < 4:
            if not analyzer.accepts_args():
                raise TypeError(
                    f'`exception_handler` should accept `4` parameters, meanwhile the given callable '
                    f'expects up to `{max_!r}`, got {function!r}.'
                )
    
    return function


def _register_exception_handler(parent, first, exception_handler):
    """
    Registers an exception handler to the respective entity. This function is wrapped inside of
    `functools.partial` with it's `parent` and `first` parameters.
    
    Parameters
    ----------
    parent : ``ExceptionHandlerInterface``
        The slasher to register the exception handler to.
    first : `bool`
        Whether the exception handler should run first.
    exception_handler : `CoroutineFunctionType`
        Exception handler to register.
    
    Returns
    -------
    exception_handler : `CoroutineFunctionType`
    
    Raises
    ------
    RuntimeError
        If `exception_handler` is given as `None`.
    """
    return parent._register_exception_handler(exception_handler, first)


class ExceptionHandlerInterface(RichAttributeErrorBaseType):
    """
    Common class for exception handleable objects.
    """
    __slots__ = ()
    
    def error(self, function = ..., *, first = False):
        """
        Registers an exception handler to the ``SlashCommandCategory``.
        
        Parameters
        ----------
        function : `None | CoroutineFunctionType`, Optional
            Function to register as exception handler.
        first : `bool` = `False`, Optional (Keyword Only)
            Whether the exception handler should run first.
        
        Returns
        -------
        exception_handler / wrapper : `CoroutineFunctionType` / `functools.partial`
            If `exception_handler` is not given, returns a wrapper.
        """
        if function is ...:
            return partial_func(_register_exception_handler, self, first)
        
        return self._register_exception_handler(function, first)
    
    
    def _register_exception_handler(self, function, first):
        """
        Registers an exception handler to the ``SlashCommandCategory``.
        
        Parameters
        ----------
        function : `CoroutineFunctionType`
            Exception handler to register.
        first : `bool`
            Whether the exception handler should run first.
        
        Returns
        -------
        exception_handler : `CoroutineFunctionType`
        """
        exception_handler = validate_exception_handler(function)
        self._store_exception_handler(exception_handler, first)
        return exception_handler
    
    
    def _store_exception_handler(self, exception_handler, first):
        """
        Stores the exception handler.
        
        Parameters
        ----------
        exception_handler : `CoroutineFunctionType`
            Exception handler to register.
        first : `bool`
            Whether the exception handler should run first.
        
        Returns
        -------
        stored : `bool`
        instance : `CoroutineFunctionType`
        """
        exception_handlers = self._exception_handlers
        if exception_handlers is None:
            exception_handlers = []
            self._exception_handlers = exception_handlers
        
        if first:
            exception_handlers.insert(0, exception_handler)
        else:
            exception_handlers.append(exception_handler)
        return True, exception_handler
    
    
    @property
    def _exception_handlers(self):
        """
        The registered exception handlers.
        
        Overwrite it as an attribute in subclasses.
        
        Returns
        -------
        _exception_handlers : `None | list<CoroutineFunctionType>`
        """
        return None
    
    
    @_exception_handlers.setter
    def _exception_handlers(self, value):
        pass



__all__ = ()

from scarletio import RichAttributeErrorBaseType

from ..opus import OpusDecoder, opus

if opus is None:
    DECODER = None
else:
    DECODER = OpusDecoder()


class VoicePacket(RichAttributeErrorBaseType):
    """
    Represents a voice packet.
    
    Attributes
    ----------
    _cache_decoded : `None | bytes`
        Cache field for ``.decoded``.
    encoded : `bytes`
        Encoded yet to decrypt data.
    """
    __slots__ = ('_cache_decoded', 'encoded')
    
    def __new__(cls, encoded):
        """
        Creates a new voice packet.
        
        Parameters
        ----------
        encoded : `bytes`
            Not yet decoded voice data.
        """
        self = object.__new__(cls)
        self.encoded = encoded
        self._cache_decoded = None
        return self
    
    
    def __repr__(self):
        """Returns the voice packet's representation."""
        repr_parts = ['<', type(self).__name__, '>']
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two packets are equal."""
        if type(self) is not type(other):
            return False
        
        return self.encoded == other.encoded
    
    
    def __hash__(self):
        """Returns the voice packet's hash value."""
        return hash(self.encoded)
    
    
    @property
    def decoded(self):
        """
        Returns the voice packet's decoded data.
        
        Returns
        -------
        data : `bytes`
        """
        decoded = self._cache_decoded
        if decoded is None:
            assert DECODER is not None
            decoded = DECODER.decode(self.encoded)
            self._cache_decoded = decoded
        
        return decoded



__all__ = ('MessageRoleSubscription',)

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_renewal, parse_subscription_listing_id, parse_tier_name, parse_total_months, put_renewal,
    put_subscription_listing_id, put_tier_name, put_total_months, validate_renewal,
    validate_subscription_listing_id, validate_tier_name, validate_total_months
)


class MessageRoleSubscription(RichAttributeErrorBaseType):
    """
    Data of a subscription purchase or renewal included with a roel subscription purchase message.
    
    Parameters
    ----------
    renewal : `bool`
        Whether the purchase is a renewal.
    subscription_listing_id : `int`
        The subscription listing's and sku's identifier to which the role subscription belongs to.
    tier_name : `str`
        The name of the tier the user subscribed to.
    total_months : `int`
        The total amount months the user has been subscribed.
    """
    __slots__ = ('renewal', 'subscription_listing_id', 'tier_name', 'total_months')
    
    def __new__(cls, *, renewal = ..., subscription_listing_id = ..., tier_name = ..., total_months = ...):
        """
        Creates a new message role subscription with the given fields.
        
        Parameters
        ----------
        renewal : `bool`, Optional (Keyword only)
            Whether the purchase is a renewal.
        subscription_listing_id : `int`, Optional (Keyword only)
            The subscription listing's and sku's identifier to which the role subscription belongs to.
        tier_name : `str`, Optional (Keyword only)
            The name of the tier the user subscribed to.
        total_months : `int`, Optional (Keyword only)
            The total amount months the user has been subscribed.
        
        Raises
        ------
        TypeError
            - Parameter of invalid type given.
        ValueError
            - Parameter of invalid value given.
        """
        # renewal
        if renewal is ...:
            renewal = False
        else:
            renewal = validate_renewal(renewal)
        
        # subscription_listing_id
        if subscription_listing_id is ...:
            subscription_listing_id = 0
        else:
            subscription_listing_id = validate_subscription_listing_id(subscription_listing_id)
        
        # tier_name
        if tier_name is ...:
            tier_name = ''
        else:
            tier_name = validate_tier_name(tier_name)
        
        # total_months
        if total_months is ...:
            total_months = 1
        else:
            total_months = validate_total_months(total_months)
        
        # construct
        self = object.__new__(cls)
        self.renewal = renewal
        self.subscription_listing_id = subscription_listing_id
        self.tier_name = tier_name
        self.total_months = total_months
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new message role subscription instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Message role subscription data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.renewal = parse_renewal(data)
        self.subscription_listing_id = parse_subscription_listing_id(data)
        self.tier_name = parse_tier_name(data)
        self.total_months = parse_total_months(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the message role subscription to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_renewal(self.renewal, data, defaults)
        put_subscription_listing_id(self.subscription_listing_id, data, defaults)
        put_tier_name(self.tier_name, data, defaults)
        put_total_months(self.total_months, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the message role subscription's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' tier_name = ')
        repr_parts.append(repr(self.tier_name))
        
        repr_parts.append(', renewal = ')
        repr_parts.append(repr(self.renewal))
        
        subscription_listing_id = self.subscription_listing_id
        if subscription_listing_id:
            repr_parts.append(', subscription_listing_id = ')
            repr_parts.append(repr(subscription_listing_id))
        
        total_months = self.total_months
        if (total_months != 1):
            repr_parts.append(', total_months = ')
            repr_parts.append(repr(total_months))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two message role subscriptions are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # renewal
        if self.renewal != other.renewal:
            return False
        
        # parse_subscription_listing_id
        if self.subscription_listing_id != other.subscription_listing_id:
            return False
        
        # tier_name
        if self.tier_name != other.tier_name:
            return False
        
        # total_months
        if self.total_months != other.total_months:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the message role subscription."""
        hash_value = 0
        
        # renewal
        hash_value ^= self.renewal
        
        # parse_subscription_listing_id
        hash_value ^= self.subscription_listing_id
        
        # tier_name
        hash_value ^= hash(self.tier_name)
        
        # total_months
        hash_value ^= self.total_months << 2
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the role subscription listing.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.renewal = self.renewal
        new.subscription_listing_id = self.subscription_listing_id
        new.tier_name = self.tier_name
        new.total_months = self.total_months
        return new
        
        
    def copy_with(self, *, renewal = ..., subscription_listing_id = ..., tier_name = ..., total_months = ...):
        """
        Copies the role subscription listing with the given fields.
        
        Parameters
        ----------
        renewal : `bool`, Optional (Keyword only)
            Whether the purchase is a renewal.
        subscription_listing_id : `int`, Optional (Keyword only)
            The subscription listing's and sku's identifier to which the role subscription belongs to.
        tier_name : `str`, Optional (Keyword only)
            The name of the tier the user subscribed to.
        total_months : `int`, Optional (Keyword only)
            The total amount months the user has been subscribed.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - Parameter of invalid type given.
        ValueError
            - Parameter of invalid value given.
        """
        # renewal
        if renewal is ...:
            renewal = self.renewal
        else:
            renewal = validate_renewal(renewal)
        
        # subscription_listing_id
        if subscription_listing_id is ...:
            subscription_listing_id = self.subscription_listing_id
        else:
            subscription_listing_id = validate_subscription_listing_id(subscription_listing_id)
        
        # tier_name
        if tier_name is ...:
            tier_name = self.tier_name
        else:
            tier_name = validate_tier_name(tier_name)
        
        if total_months is ...:
            total_months = self.total_months
        else:
            total_months = validate_total_months(total_months)
        
        # Construct
        new = object.__new__(type(self))
        new.renewal = renewal
        new.subscription_listing_id = subscription_listing_id
        new.tier_name = tier_name
        new.total_months = total_months
        return new



__all__ = ('StickerFormat', 'StickerType', )

from scarletio import export

from ...bases import Preinstance as P, PreinstancedBase


@export
class StickerFormat(PreinstancedBase, value_type = int):
    """
    Represents a message sticker's format's type.
    
    Attributes
    ----------
    extension : `str`
        The extension of the sticker format type.
    
    name : `str`
        The name of the message sticker format type.
    
    value : `int`
        The Discord side identifier value of the message sticker format type.
    
    Type Attributes
    ---------------
    Every predefined sticker format type can be accessed as type attribute as well:
    
    +-----------------------+-----------+-------+---------------+
    | Type attribute name   | name      | value | extension     |
    +=======================+===========+=======+===============+
    | none                  | none      | 0     | png           |
    +-----------------------+-----------+-------+---------------+
    | png                   | png       | 1     | png           |
    +-----------------------+-----------+-------+---------------+
    | apng                  | apng      | 2     | png           |
    +-----------------------+-----------+-------+---------------+
    | lottie                | lottie    | 3     | json          |
    +-----------------------+-----------+-------+---------------+
    | gif                   | gif       | 4     | gif           |
    +-----------------------+-----------+-------+---------------+
    """
    __slots__ = ('extension', )
    
    def __new__(cls, value, name = None, extension = 'png'):
        """
        Creates a new sticker format type.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the message sticker format type.
        
        name : `None | str` = `None`, Optional
            The name of the message sticker format type.
        
        extension : `str` = `'png'`, Optional
            The extension of the sticker format type.
        """
        self = PreinstancedBase.__new__(cls, value, name)
        self.extension = extension
        return self
    
    
    # predefined
    none = P(0, 'none', 'png')
    png = P(1, 'png', 'png')
    apng = P(2, 'apng', 'png')
    lottie = P(3, 'lottie', 'json')
    gif = P(4, 'gif', 'gif')


class StickerType(PreinstancedBase, value_type = int):
    """
    Represents a message sticker's type.
    
    Attributes
    ----------
    value : `int`
        The Discord side identifier value of the message sticker type.
    
    name : `str`
        The name of the message sticker type.
    
    Type Attributes
    ---------------
    Every predefined sticker type can be accessed as type attribute as well:
    
    +-----------------------+-----------+-------+
    | Type attribute name   | name      | value |
    +=======================+===========+=======+
    | none                  | none      | 0     |
    +-----------------------+-----------+-------+
    | standard              | standard  | 1     |
    +-----------------------+-----------+-------+
    | guild                 | guild     | 2     |
    +-----------------------+-----------+-------+
    """
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    standard = P(1, 'standard')
    guild = P(2, 'guild')
